""" Utilties to interface gwsurrogate models with the generic PyCBC
gw waveform interfaces
"""
import gwsurrogate as gws

_cached_models = {}

def gws_td_gen(**params):
    """ Generate pycbc format time domain GW polarizations

    Parameters
    ----------
    params: dict
        A dictionary containing the waveform parameters

    Returns
    -------
    hp: pycbc.types.TimeSeries
        The time series containg the plus polarization in the
        radiation frame.
    hc: pycbc.types.TimeSeries
        The time series containg the cross polarization in the
        radiation frame.
    """
    import gwsurrogate as gws
    from pycbc.types import TimeSeries
    from pycbc import conversions as conv

    # Parameter names should follow the convention of
    # https://github.com/gwastro/pycbc/blob/master/pycbc/waveform/parameters.py#L162
    # but are not limited to these names for additional effects
    apx = params['approximant'].replace('GWS-', '')

    if apx not in _cached_models:
        _cached_models[apx] = gws.LoadSurrogate(apx)
    model = _cached_models[apx]

    # convert from pycbc parameter name conventions to gwsurrogate
    q1 = params['mass1'] / params['mass2']
    q2 = params['mass2'] / params['mass1']


    q = [q1, q2]
    chi = [[params['spin1x'], params['spin1y'], params['spin1z']],
           [params['spin2x'], params['spin2y'], params['spin2z']]]

    # sort so first mass is larger
    if q1 >= 1:
        a, b = 0, 1
    else:
        a, b = 1, 0

    tides = None
    if model.keywords['Tidal']:
        lam = [params['lambda1'], params['lambda2']]
        tides = {'Lambda1':lam[a], 'Lambda2':lam[b]}
        tides = {k:tides[k] if tides[k] is not None else 0.0 for k in tides}

    M = params['mass1'] + params['mass2']
    if params['f_ref'] == 0:
        params['f_ref'] = params['f_lower']

    _, h, _ = model(q[a], chi[a], chi[b],
                    M=M,
                    dist_mpc=params['distance'],
                    dt=params['delta_t'],
                    phi_ref=params['coa_phase'],
                    f_ref=params['f_ref'],
                    f_low=params['f_lower'],
                    tidal_opts=tides,
                    inclination=params['inclination'],
                    units='mks')

    # Define time=0 as the waveform peak to be consistent with
    # convention
    peak_time = abs(h).argmax() * params['delta_t']
    hp = TimeSeries(h.real, delta_t=params['delta_t'], epoch=-peak_time)
    hc = TimeSeries(h.imag, delta_t=params['delta_t'], epoch=-peak_time)
    return hp, hc



""" download all models to be tested in test_model_regression.py

This is useful to do when using continuous integration. """

import gwsurrogate as gws
import hashlib, os


def md5(fname):
  """ Compute has from file. code taken from
  https://stackoverflow.com/questions/3431825/generating-an-md5-checksum-of-a-file"""
  hash_md5 = hashlib.md5()
  with open(fname, "rb") as f:
    for chunk in iter(lambda: f.read(4096), b""):
      hash_md5.update(chunk)
  return hash_md5.hexdigest()

models = ['SpEC_q1_10_NoSpin_linear_alt',
'NRHybSur3dq8',
'NRHybSur3dq8_CCE',
'SpEC_q1_10_NoSpin',
'SpEC_q1_10_NoSpin_linear',
'NRSur7dq4',
'NRHybSur2dq15'
]

for model in models:
  print("Downloading model %s ..."%model)
  gws.catalog.pull(model)
  surr_url = gws.catalog._surrogate_world[model].url
  path_to_model = gws.catalog.download_path()+os.path.basename(surr_url)
  print("md5 Hash of %s is %s"%(model,md5(path_to_model)))
  if not gws.catalog.is_file_recent(path_to_model):
     print("File download failed!") 
     assert(False)



"""
Tests all relevant data generated by tutorial notebooks.
Regression data is stored in npz files.
"""

from __future__ import division
import numpy as np
import gwsurrogate as gws
import os, sys

# set global tolerances for floating point comparisons (see np.testing.assert_allclose)
atol = 0.0
rtol = 1.e-11

# try importing data. If it doesn't exist, download it
try:
  reg_data = np.load('test/gws_regression_data_v2/data_notebook_basics_lesson1_v2.npz')
except:
  print("Downloading regression data...")
  #os.system('wget --directory-prefix=test https://www.dropbox.com/s/07t84cpmmqjya69/gws_regression_data.tar.gz')
  os.system('wget --directory-prefix=test https://www.dropbox.com/s/2z8xvrdu4kdjc18/gws_regression_data_v2.tar.gz')
  os.system('tar -xf test/gws_regression_data_v2.tar.gz -C test/')

path_to_surrogate = \
'tutorial/TutorialSurrogate/EOB_q1_2_NoSpin_Mode22/l2_m2_len12239M_SurID19poly/'
EOBNRv2_sur = gws.EvaluateSingleModeSurrogate(path_to_surrogate)


def test_notebook_basics_lesson1():
  """ Regression test data from notebook example.

      Data created on 5/30/2015 with

np.savez('data_notebook_basics_lesson1.npz',t=t,hp=hp,hc=hc,amp=amp,phase=phase,phi_m=phi_m,h_adj=h_adj)

      After gwtools changes to constants (8/24/2018), _v2 of this data was created"""

  # because M and dist are provided, a physical GW is generated
  t, hp, hc  = EOBNRv2_sur(q=1.7, M=80.0, dist=1.0, phi_ref = 0.0, f_low = 10.0)
  amp, phase = EOBNRv2_sur.amp_phase(hp + 1j*hc)

  phi_m = EOBNRv2_sur.phi_merger(hp + 1j*hc)
  h_adj = EOBNRv2_sur.adjust_merger_phase(hp + 1j*hc,2.0)

  # load regression data
  reg_data = np.load('test/gws_regression_data_v2/data_notebook_basics_lesson1_v2.npz')

  np.testing.assert_allclose(t,reg_data['t'], rtol=rtol, atol=atol)
  np.testing.assert_allclose(hp,reg_data['hp'], rtol=rtol, atol=atol)
  #np.testing.assert_allclose(hc,reg_data['hc'], rtol=rtol, atol=atol)
  # one value is almost zero, causes test to fail
  np.testing.assert_allclose(hc+1.e-15,reg_data['hc']+1.e-15, rtol=rtol, atol=atol)
  np.testing.assert_allclose(amp,reg_data['amp'], rtol=rtol, atol=atol)
  np.testing.assert_allclose(phase,reg_data['phase'], rtol=rtol, atol=atol)
  np.testing.assert_allclose(phi_m,reg_data['phi_m'], rtol=rtol, atol=atol)
  np.testing.assert_allclose(h_adj,reg_data['h_adj'], rtol=rtol, atol=atol)

def test_notebook_basics_lesson2():
  """ Regression test data from notebook example.

      Data created on 5/30/2015 with

np.savez('data_notebook_basics_lesson2.npz',t_resamp=t_resamp,hp_resamp=hp_resamp,hc_resamp=hc_resamp)

      After gwtools changes to constants (8/24/2018), _v2 of this data was created"""

  #t, hp, hc = EOBNRv2_sur(q=1.2)
  t_resamp, hp_resamp, hc_resamp = \
    EOBNRv2_sur(1.2,times=np.linspace(EOBNRv2_sur.tmin-1000,EOBNRv2_sur.tmax+1000,num=3000))

  # load regression data
  reg_data = np.load('test/gws_regression_data_v2/data_notebook_basics_lesson2_v2.npz')

  np.testing.assert_allclose(t_resamp,reg_data['t_resamp'], rtol=rtol, atol=atol)
  np.testing.assert_allclose(hp_resamp,reg_data['hp_resamp'], rtol=rtol, atol=atol)
  np.testing.assert_allclose(hc_resamp,reg_data['hc_resamp'], rtol=rtol, atol=atol)


def test_notebook_basics_lesson3():
  """ regression test data from notebook example.

      Data created on 5/30/2015 with

np.savez('data_notebook_basics_lesson3_v2.npz',eim_pts=eim_pts, T_eim=T_eim,
          greedy_pts=greedy_pts,tmin=EOBNRv2_sur.tmin,tmax=EOBNRv2_sur.tmax,
          dt=EOBNRv2_sur.dt,tunits=EOBNRv2_sur.t_units)

      After gwtools changes to constants (8/24/2018), _v2 of this data was created"""


  t, hp, hc  = EOBNRv2_sur(1.7,80.0,1.0)
  eim_pts    = EOBNRv2_sur.eim_indices
  T_eim      = t[eim_pts]
  greedy_pts = EOBNRv2_sur.greedy_points

  tunits = EOBNRv2_sur.t_units
  dt = EOBNRv2_sur.dt
  tmax = EOBNRv2_sur.tmax
  tmin = EOBNRv2_sur.tmin


  reg_data = np.load('test/gws_regression_data_v2/data_notebook_basics_lesson3_v2.npz')

  np.testing.assert_allclose(eim_pts,reg_data['eim_pts'], rtol=rtol, atol=atol)
  np.testing.assert_allclose(T_eim,reg_data['T_eim'], rtol=rtol, atol=atol)
  np.testing.assert_allclose(greedy_pts,reg_data['greedy_pts'], rtol=rtol, atol=atol)
  np.testing.assert_allclose(dt,reg_data['dt'], rtol=rtol, atol=atol)
  np.testing.assert_allclose(tmax,reg_data['tmax'], rtol=rtol, atol=atol)
  np.testing.assert_allclose(tmin,reg_data['tmin'], rtol=rtol, atol=atol)
  assert(str(tunits) == str(reg_data['tunits']))


def test_notebook_basics_lesson4():
  """ Regression test data from notebook example.

      Data created on 5/30/2015 with

np.savez('data_notebook_basics_lesson4.npz',times=times,b_5=b_5,e_5=e_5,h_5=h_5,h_5_surr=h_5_surr)

      After gwtools changes to constants (8/24/2018), _v2 of this data was created"""


  b_5   = EOBNRv2_sur.basis(4,'cardinal')
  e_5   = EOBNRv2_sur.basis(4,'orthogonal')
  h_5   = EOBNRv2_sur.basis(4,'waveform')
  junk, hp_5_surr, hc_5_surr = EOBNRv2_sur(EOBNRv2_sur.greedy_points[4])
  nrm_5 = EOBNRv2_sur.norm_eval(EOBNRv2_sur.greedy_points[4])
  hp_5_surr = hp_5_surr / nrm_5 
  hc_5_surr = hc_5_surr / nrm_5
  h_5_surr = hp_5_surr + 1j*hc_5_surr

  # load regression data
  reg_data = np.load('test/gws_regression_data_v2/data_notebook_basics_lesson4_v2.npz')

  np.testing.assert_allclose(b_5,reg_data['b_5'], rtol=rtol, atol=atol)
  np.testing.assert_allclose(e_5,reg_data['e_5'], rtol=rtol, atol=atol)
  np.testing.assert_allclose(h_5,reg_data['h_5'], rtol=rtol, atol=atol)
  np.testing.assert_allclose(h_5_surr,reg_data['h_5_surr'], rtol=rtol, atol=atol)

  # basis orthogonality
  dt  = 1.0/2048.0 # found from surrogate *.dat file
  e_6 = EOBNRv2_sur.basis(5,'orthogonal')
  orthogonal_test = np.sum(e_5*np.conj(e_6)) * dt
  normalized_test = np.sum(e_5*np.conj(e_5)) * dt
  np.testing.assert_allclose(orthogonal_test, 0, rtol=0, atol=1.e-14)
  np.testing.assert_allclose(normalized_test, 1.0, rtol=0, atol=1.e-14)

#def test_notebook_basics_lesson5():
#
#  assert(False) #TODO: code me (download surrogates need to be rebuilt)

def test_tPNT2Tidal():
  out = gws.new.tidal_functions.PNT2Tidal(1.e-2, 2.0, 100., 1000. , .3, -.7, 100., 1000., .2, .5)
  assert( np.abs(out[0] + 15351562.5216222 )/ 15351562.5216222 < 1.e-12)
  assert( np.abs(out[1] + 61.40625000865016 )/ 61.40625000865016 < 1.e-12)

# when regression tests fail, this information is useful
def test_package_versions():
  print("\n============= VERSIONS ============", file=sys.stdout)
  print("numpy version = %s"%np.__version__, file=sys.stdout)
  import sklearn
  sklearn.__version__
  print("sklearn version = %s"%sklearn.__version__, file=sys.stdout)
  print("\n", file=sys.stdout)
 
if __name__ == '__main__':
  print("Running test 1...")
  test_notebook_basics_lesson1()
  print("Running test 2...")
  test_notebook_basics_lesson2()
  print("Running test 3...")
  test_notebook_basics_lesson3()
  print("Running test 4...")
  test_notebook_basics_lesson4()




"""
Simple unit tests.
"""

from __future__ import division
import numpy as np
import gwsurrogate as gws
import os

path_to_surrogate = 'tutorial/TutorialSurrogate/EOB_q1_2_NoSpin_Mode22/'

def test_orbital_symmetry_flags():
  # TODO: add 4d2s -- this will give non-trival combinations of this flag
  """ Check valid combinations of orbital symmetry flags"""

  EOBNRv2_sur = gws.EvaluateSurrogate(path_to_surrogate)

  # These three combinations should pass
  EOBNRv2_sur = gws.EvaluateSurrogate(path_to_surrogate, use_orbital_plane_symmetry=True)
  modes, t, hp, hc = EOBNRv2_sur(q=1.14,ell=[2],m=[2],mode_sum=False,fake_neg_modes=True)

  EOBNRv2_sur = gws.EvaluateSurrogate(path_to_surrogate, use_orbital_plane_symmetry=True)
  modes, t, hp, hc = EOBNRv2_sur(q=1.14,ell=[2],m=[2],mode_sum=False,fake_neg_modes=False)

  EOBNRv2_sur = gws.EvaluateSurrogate(path_to_surrogate, use_orbital_plane_symmetry=False)
  modes, t, hp, hc = EOBNRv2_sur(q=1.14,ell=[2],m=[2],mode_sum=False,fake_neg_modes=False)

  # This should fail with a ValueError. So seeing this error is a passed test
  try:
    EOBNRv2_sur = gws.EvaluateSurrogate(path_to_surrogate, use_orbital_plane_symmetry=False)
    modes, t, hp, hc = EOBNRv2_sur(q=1.14,ell=[2],m=[2],mode_sum=False,fake_neg_modes=True)
  except ValueError:
    pass



""" catalog of parametric fitting functions """

from __future__ import division # for python 2

__copyright__    = "Copyright (C) 2014 Scott Field and Chad Galley"
__email__        = "sfield@umassd.edu, crgalley@tapir.caltech.edu"
__status__       = "testing"
__author__       = "Jonathan Blackman, Scott Field, Chad Galley"
__contributors__ = []

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

import numpy as np
from gwtools import gwtools as gwtools # from the package gwtools, import the module gwtools (gwtools.py)....
from scipy.interpolate import splev


#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
def polyval_1d(coeffs,x):
  """ 1D polynomial defined by coeffs vector and evaluated 
  as numpy.polyval(coeffs,x)"""
  return np.polyval(coeffs, x)


#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
def spline_1d(coeffs,x):
  """ 1d spline defined by knots and spline coeffs. """
  return splev(x, coeffs)

#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
def ampfitfn1_1d(coeffs,x):
  """ PN inspired ampitude fit a0 + a1*nu**a2"""

  a0, a1, a2 = coeffs[:3]
  nu = gwtools.q_to_nu(x)
  return a0 + a1*nu**a2

#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
def ampfitfn2_1d(coeffs, x):
  """ PN inspired amplitude fit a0 + a1*np.abs(0.25-nu)**0.5 + a2*np.log(nu/0.25)"""

  a0, a1, a2 = coeffs[:3]
  nu = gwtools.q_to_nu(x)
  return a0 + a1*np.abs(0.25-nu)**0.5 + a2*np.log(nu/0.25)

#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
def phifitfn1_1d(coeffs,x):
  """ PN inspired phase fit a0 + a1*nu + a2*nu**2 + a3*np.log(nu)"""

  a0, a1, a2, a3 = coeffs[:4]
  nu = gwtools.q_to_nu(x)
  return a0 + a1*nu + a2*nu**2 + a3*np.log(nu)

#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
def ampfitfn3_1d(coeffs, x):
  """ PN inspired amplitude fit plus polynomial,
      a0*np.sqrt(1 - x) + a1*np.log(x) + a2(1 - x) + ... + aN*(1-x)**(N-1)"""

  a0 = coeffs[-1]
  a1 = coeffs[-2]

  polyCoefs = [c for c in coeffs[:-2]]
  polyCoefs.append(0.)

  return a0*np.sqrt(1. - x) + a1*np.log(x) + np.polyval(polyCoefs,1. - x)

#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
def ampfitfn4_1d(coeffs, x):
  """ a0*sqrt(1.0-x) + a1*(1.0-x) + a2*(1.0-x)^2 + a3*(1.0-x)^3"""

  a0, a1, a2, a3 = coeffs[:4]
  return a0*np.sqrt(1.0-x) + a1*(1.0-x) + a2*np.power(1.0-x,2) + a3*np.power(1.0-x,3)

#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
def ampfitfn5_1d(coeffs,x):
  """a0*np.sqrt(1 - x) + a1(1 - x) + ... + aN*(1-x)**N"""
  a0 = coeffs[-1]
  polyCoefs = [c for c in coeffs[:-1]]
  polyCoefs.append(0.)
  
  return a0*np.sqrt(1. - x) + np.polyval(polyCoefs,1. - x)

#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
def ampfitfn6_1d(coeffs,x):
  """a0*np.sqrt(1 - x) + a1*(1 - x)**1.5 + [ a2*(1 - x) + ... + aN*(1-x)**(N-1)"""
  a0 = coeffs[-1]
  a1 = coeffs[-2]
  polyCoefs = [c for c in coeffs[:-2]]
  polyCoefs.append(0.)

  return a0*np.sqrt(1. - x) + a1*(1. - x)**1.5 + np.polyval(polyCoefs,1. - x)

def emri_normalization_logq(coeffs,x):
  """ Teukolsky solver that generates EMRI data normalizes by q, since 
  the EMRI surrogates use log(q) parameterizaiton, we need to undo
  the log and apply the q-dependent normalization. No coefficients are
  needed, but we keep this structure for common API. """

  q = np.exp(x)
  return 1.0/q

def BHPT_normalization_log10q(coeffs,x):
  """ Teukolsky solver that generates EMRI data normalizes by q, since 
  the BHPT surrogates use log10(q) parameterizaiton, we need to undo
  the log and apply the q-dependent normalization. No coefficients are
  needed, but we keep this structure for common API. """

  q = 10**x
  return 1.0/q

### these are for switching from (q,M) to surrogate's parameterization ###
def q_to_q(q):
  """ identity map from q to q
  
  Surrogates with this parameterization expect its user intput 
  to be the mass ratio q, and "map" to the internal surrogate's 
  parameterization which is also q
  
  The surrogates training interval is in mass ratio
  """
  return q

def q_to_nu(q):
  """ map from q to symmetric mass ratio nu
  
  Surrogates with this parameterization expect its user intput 
  to be the mass ratio q. 
  
  The surrogate will map q to the internal surrogate's 
  parameterization which is nu
  
  The surrogates training interval is quoted in symmetric mass ratio.
  """
  return gwtools.q_to_nu(q)

def q_to_logq(q):
  """ map from q to log(q)
  
  Surrogates with this parameterization expect its user intput 
  to be the mass ratio q. 
  
  The surrogate will map q to the internal surrogate's 
  parameterization which is log(q)
  
  The surrogates training interval is quoted in log(q).
  """
  return np.log(q)

def q_to_log10q(q):
  """ map from q to log10(q)
  
  Surrogates with this parameterization expect its user intput 
  to be the mass ratio q. 
  
  The surrogate will map q to the internal surrogate's 
  parameterization which is log10(q)
  
  The surrogates training interval is quoted in log10(q).
  """
  return np.log10(q)

def BHPT_nrcalib_functional_form(x,a,b,c,d):
    """
    functional form of alpha and beta scaling factors in BHPTNRSur1dq1e4
    for NR calibration in the comparable mass regime
    """
    return 1 + a*x + b*x**2 + c*x**3 + d*x**4

#-
### dictionary of fitting functions ###
function_dict = {
                 "polyval_1d": polyval_1d,
                 "spline_1d": spline_1d,
                 "ampfitfn1_1d": ampfitfn1_1d,
                 "ampfitfn2_1d": ampfitfn2_1d,
                 "ampfitfn4_1d": ampfitfn4_1d,
                 "phifitfn1_1d": phifitfn1_1d,
                 "nuSingularPlusPolynomial": ampfitfn5_1d,
                 "nuSingular2TermsPlusPolynomial": ampfitfn6_1d,
                 "emri_normalization_logq":emri_normalization_logq,
                 "BHPT_normalization_log10q": BHPT_normalization_log10q,
                 "BHPT_nrcalib_functional_form": BHPT_nrcalib_functional_form,
                 "q_to_q": q_to_q,
                 "q_to_nu": q_to_nu,
                 "q_to_logq": q_to_logq,
                 "q_to_log10q": q_to_log10q
                 }



"""
When adding a new model, its good to also setup a model regression test.

Here are the steps for doing this (the last step is running this script)


First make sure the current tests run on your laptop
====================================================

Download all model data

>>>  python download_regression_models.py  # run from tests folder

The run all of the tests

>>> pytest -v -s 

This will also download a file called model_regression_data.h5,
which will be used later on.


Setup tests for the new model (after you have implemented the model)
====================================================================
1) Add your model to download_regression_models.py
2) Add model to interface list in test_model_regression.py
3) Write a routine to decide at which points to sample the model in test_model_regression.py


Generate new model regression data
===================================
1) Generate the new regression data by running

>>> python test_model_regression.py  #run from tests folder

which will produce a file called model_regression_data_new.h5. This file will contain
regression data for both your new model as well as previous models. However, we
don't want to overwrite previous model regression data. 

2) The last step is to copy your model-specific regression data into
the existing file model_regression_data.h5. This way we don't overwrite the previous 
regression data.

>>> python add_model_regression_data.py

After running this step, model_regression_data.h5 should have the new model's
regression data. 

3) model_regression_data_new.h5 can be deleted
4) run check_model_regression_data.py to confirm that your data has been 
correctly added and the old data is unchanged.
"""

import h5py
import numpy as np
import shutil

original_regression_data = "model_regression_data.h5"
extra_regression_data    = "model_regression_data_new.h5"
new_model = "NRHybSur3dq8_CCE" # name of new model with regression data in fp_new but not fp_old

shutil.copyfile(original_regression_data, "model_regression_data-original.h5")


fp_new_data = h5py.File(extra_regression_data,'r')
fp_old_data = h5py.File(original_regression_data,'r+')


for model in fp_new_data.keys():
    if model in fp_old_data.keys():
        print("model %s already exists in regression dataset. Skipping...."%model)
    else:
        print("model %s does not exist in regression dataset. Adding...."%model)
        fp_new_data.copy(model,fp_old_data)


fp_new_data.close()
fp_old_data.close()


# the file model_regression_data should have the new model regression data for new_model
# we now check that the copying went OK.
# 
# Check 1: the new_model is in the new datafile but not the old one
# Check 2: the original data is unchanged

## --- test that the copying worked correctly --- ##


fp_new = h5py.File("model_regression_data.h5")
fp_old = h5py.File("model_regression_data-original.h5",'r')


print("models in new datafile")
for model in fp_new.keys(): # loop over models
    print(model)
print("---------------")
print("models in old datafile")
for model in fp_old.keys(): # loop over models
    print(model)

for model in fp_new.keys(): # loop over models
    if model != new_model:
        model_regession_data = fp_new[model]
        for paramter in model_regession_data.keys(): # loop over parameters
            model_regession_data_i = model_regession_data[paramter]
            for data in model_regession_data_i.keys(): # loop over data for each parameter
                if data != "parameter" and data != "time":
                    x = '/'+model+'/'+paramter+'/'+data
                    y1 = model_regession_data_i[data][:,:]
                    y2 = fp_old[x][:,:]
                    y = y1 - y2
                    if np.max(np.abs(y)) == 0.0:
                        pass
                    else:
                        print("New and old regression data doesn't agree.")
                        print(x)
                        print(y)
    else:
        print("model %s doesn't exist in original dataset"%new_model)



"""
After adding a new model's regression data (see add_model_regression_data.py) 

run this script to check that the common hdf5 groups in the new and old hdf5
files are the same. 
"""

import h5py
import numpy as np

# new and old hdf5 regression files
fp_old = h5py.File("model_regression_data-original.h5")
fp_new = h5py.File("model_regression_data.h5",'r')

# name of new model with regression data in fp_new but not fp_old
# this will exclude NEW_MODEL from the check
# set this to a nonsenese name if we want to check all models
new_model = "NRHybSur3dq8_CCE"


# Visual check: all enteries are the same except the new one added to fp_new
print("models in new datafile")
for model in fp_new.keys(): # loop over models
    print(model)
print("\n\n---------------\n\n")
print("models in old datafile")
for model in fp_old.keys(): # loop over models
    print(model)


for model in fp_new.keys(): # loop over models
    if model != new_model: # Skip new model we just added
        model_regession_data = fp_new[model]
        for paramter in model_regession_data.keys(): # loop over parameters
            model_regession_data_i = model_regession_data[paramter]
            for data in model_regession_data_i.keys(): # loop over data for each parameter
                if data != "parameter" and data != "time":
                    x = '/'+model+'/'+paramter+'/'+data
                    #print(x)
                    y1 = model_regession_data_i[data][:,:]
                    y2 = fp_old[x][:,:]
                    y = y1 - y2
                    #print( np.max(np.abs(y1)) )
                    if np.max(np.abs(y)) == 0.0:
                        pass
                    else:
                        print("New and old regression data doesn't agree.")
                        print(x)
                        print(y)



""" Gravitational Wave Surrogate classes for text and hdf5 files"""

from __future__ import division  # for py2

__copyright__ = "Copyright (C) 2014 Scott Field and Chad Galley"
__email__     = "sfield@astro.cornell.edu, crgalley@tapir.caltech.edu"
__status__    = "testing"
__author__    = "Kevin Barkett"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

# adding "_" prefix to potentially unfamiliar module names
# so they won't show up in gws' tab completion
import numpy as np
from scipy.special import fresnel


################## NS Universal Relations

def UniversalRelationLambda2ToOmega2(lambda2):
  """ arXiv:1408.3789 eqn 3.5 with coeffs from Table I

   Input  -- ell=2 dimensionless tidal deformability lambda2 = 2/3*k2*C^5
   Output -- M_NS * omega2, the dimensionless frequency """

  if(lambda2<0):
    raise ValueError("ERROR: lambda2 value must be positive!")
  eta = np.log(lambda2)
  p = [0.182, -6.836e-3, -4.196e-3, 5.215e-4, -1.857e-5]
  omega2 = p[0] + eta*(p[1]+eta*(p[2]+eta*(p[3]+eta*(p[4]))))
  return omega2

def UniversalRelationLambda3ToOmega3(lambda3):
  """ arXiv:1408.3789 eqn 3.5 with coeffs from Table I

   Input  -- ell=3 dimensionless tidal deformability lambda3 = 2/15*k3*C^7
   Output -- M_NS * omega3, the dimensionless frequency """

  if(lambda3<0):
    raise ValueError("ERROR: lambda3 value must be positive!")
  eta = np.log(lambda3)
  p = [2.245e-1, -1.500e-2, -1.412e-3, 1.832e-4, -5.561e-6]
  omega3 = p[0] + eta*(p[1]+eta*(p[2]+eta*(p[3]+eta*(p[4]))))
  return omega3

def UniversalRelationLambda2ToLambda3(lambda2):
  """ arXiv:1311.0872 eqn 60 with coeffs from Table I

   Input  -- ell=2 dimensionless tidal deformability lambda2 = 2/3*k2*C^5
   Output -- ell=3 dimensionless tidal deformability lambda3 = 2/15*k3*C^7 """

  if(lambda2<0):
    raise ValueError("ERROR: lambda2 value must be positive!")
  eta = np.log(lambda2)
  p = [-1.15, 1.18, 2.51e-2, -1.31e-3, 2.52e-5]
  lambda3 = np.exp(p[0] + eta*(p[1]+eta*(p[2]+eta*(p[3]+eta*(p[4])))))
  return lambda3

def UniversalRelationLambda2ToAqm(lambda2):
  """ arXiv:1303.1528 eqn 54 with coeffs from Table I

   Input  -- ell=2 dimensionless tidal deformability lambda2 = 2/3*k2*C^5
   Output -- dimensionless rotationally-induced quadrupole moment (eqn 40)
             Aqm = -Q/(M^3 chi^2) """

  if(lambda2<0):
    raise ValueError("ERROR: lambda2 value must be positive!")
  eta = np.log(lambda2)
  p = [0.194, 0.0936, 0.0474, -4.21e-3, 1.23e-4]
  Aqm = np.exp(p[0] + eta*(p[1]+eta*(p[2]+eta*(p[3]+eta*(p[4])))))
  return Aqm

def UniversalRelationLambda2ToI(lambda2):
  """arXiv:1303.1528 eqn 54 with coeffs from Table I
   Input  -- ell=2 dimensionless tidal deformability lambda2 = 2/3*k2*C^5
   Output -- dimensionless moment of inertia (eqn 26); I = \bar{I}*M^3"""

  if(lambda2<0):
    raise ValueError("ERROR: lambda2 value must be positive!")
  eta = np.log(lambda2)
  p = [1.47, 0.0817, 0.0149, 2.87e-4, -3.64e-5]
  Aqm = np.exp(p[0] + eta*(p[1]+eta*(p[2]+eta*(p[3]+eta*(p[4])))))
  return Aqm

################## Tidal deformability functions

def EffectiveDeformabilityFromDynamicalTides(orbital_freq,omega_fmode,ell_mode,q):
  """arxiv:1608.01907, eqn 6.51

   Inputs:
     orbital_freq -- orbital frequency of the binary
     omega_fmode  -- f-mode dimensionless angular frequency of the ell-polar
                     mode of the deformed object
     ell_mode     -- ell-polar deformability (2 = quadrupolar; 3 = octopolar)
     q            -- mass ratio
   Outputs:
     The effective amplification of the ell-polar tidal deformability as a
     function of the oribital frequency """

  if(ell_mode not in [2,3]):
    raise ValueError("ERROR: 'ell_mode' only implemented for ell=2 or ell=3!")
  
  # If the input to the universal relations is too small (lambda2<1), then the
  #   fits can return unphysical negative resonance frequencies, in such cases
  #   just assume there is no significant resonance
  if(omega_fmode<=0):
    return np.ones(len(orbital_freq))

  # resonanace => freq_ratio == 1
  freq_ratio = omega_fmode / ell_mode / orbital_freq

  # Intermediate values
  # Because it only matters what the symmetric mass ratio is, XA*XB, it doesn't
  #   matter whether the input is q or 1/q
  X1 = q/(q+1.)
  X2 = 1. - X1
  epsilon = 256./5.*X1*X2*pow(omega_fmode/ell_mode,5./3.)
  t_hat = 8./5./np.sqrt(epsilon)*(1 - pow(freq_ratio,5./3.))
  omega_prime = 3./8.

  # Fresnel integral values
  ssa, csa = fresnel(t_hat*np.sqrt(2*omega_prime/np.pi))

  betaDTOverA = (
          pow(omega_fmode,2)/(pow(omega_fmode,2)-pow(ell_mode*orbital_freq,2))
          + pow(freq_ratio,2)/2./np.sqrt(epsilon)/t_hat/omega_prime
          + pow(freq_ratio,2)/np.sqrt(epsilon)*np.sqrt(np.pi/8./omega_prime)
          * (np.cos(omega_prime*t_hat*t_hat)*(1+2.*ssa)
            -np.sin(omega_prime*t_hat*t_hat)*(1+2.*csa))
          )
  # When 'orbital_freq' is close enough to 0, this formula seems to be
  #   numerically unstable though the asympotic limit should be 1. So when small
  #   enough, just set it to that value
  # NOTE: this cutoff value should match value used in
  #   EffectiveDissipativeDynamicalTides
  omega_cutoff = 5.e-5
  betaDTOverA[abs(orbital_freq)<=omega_cutoff] = 1.

  if ell_mode==2:
    return (1.+3.*betaDTOverA)/4.
  else:
    return (3.+5.*betaDTOverA)/8.

def EffectiveDissipativeDynamicalTides(orbitalFreq,effDefA,omegafModeA,XA):
  """ arXiv:1702.02053, eqn 15
   NOTE: there is an error in the paper that is correct in the XLALSim code!!
   This returns the effective correction to the ell=2 (quadrupole)
   deformability which used to compute the strain correction due to tidal
   effects """

  # NOTE: in the paper the input mass fraction is of the companion object, not
  # the object being deformed
  XB = 1.-XA
  effDissDyn = ((omegafModeA*omegafModeA*(effDefA-1.)
          + 6.*effDefA*XB*orbitalFreq*orbitalFreq)
        / (3.*orbitalFreq*orbitalFreq*(1.+2.*XB)))

  # When 'orbital_freq' is close enough to 0, this formula seems to be
  #   numerically unstable though the asympotic limit should be 1. So when small
  #   enough, just set it to that value
  # NOTE: this cutoff value should match value used in
  #   EffectiveDeformabilityFromDynamicalTides
  omega_cutoff = 5.e-5
  effDissDyn[abs(orbitalFreq)<=omega_cutoff] = 1.

  return effDissDyn


def Beta22_1(X):
  """ Coefficient for the 1PN tidal correction to h_22 mode
   Defined in arXiv:1203.4352, eqn A15
   Input
     X -- mass fraction of object being tidally deformed
   Output
     double, coeff for 1PN tidal correction to h_22 mode """ 

  return (-202.+560.*X-340.*X*X+45.*X*X*X)/42./(3.-2.*X)

def StrainTidalEnhancementFactor(lll,mmm,qqq,lambdaA,lambdaB,v):
  """ Strain enhancement by mode for the quadrupole (ell=2) love number
   Static correction is arXiv:1203.4352, eqn A14-A17
   Effective dynamical tide correction is in arXiv:1702.02053, eqn 15

   Because the strain correction is just an amplitude correction (specifically
     an amplification because the tidal effects enhance the GW radiation), just
     need to return the amplitude of the correction and ignore the phase of the
     mode

   If using effective tidal deformability from dynamic tides, must be included
     into lambdaA,lambdaB before passing them into this function """

  # mass fractions
  XA = qqq/(1.+qqq)
  XB = 1.-XA
  # Only beta22_1 is known, so set all others to 0, include them later into
  #   the formula when they are figured out
  beta22_1A = Beta22_1(XA)
  beta22_1B = Beta22_1(XB)
  #beta22_2A = beta22_2B = 0.
  #beta21_1A = beta21_1B = 0.
  #beta33_1A = beta33_1B = 0.
  #beta31_1A = beta31_1B = 0.

  # eta = symmetric mass ratio, MA*MB/M/M, XA*XB
  # delta = XA-XB
  eta = XA*XB

  v2 = v*v
  v4 = v2*v2
  v10 = v4*v4*v2

  # The leading part of the Newtonian expression, from arXiv:1310.1528, eqn 327
  # The formula used his is dimensionless with the M/R factor scaled out
  # (looks like there is a typo where the 'm' in the leading term is supposed
  # to be mass M?)
  newt = (8*eta*np.sqrt(np.pi/5))*v2
  newtlm = tidelm = 0

  # The kappaA_ell used in 1203.4352 is different from the dimensionless
  # lambdaA normally used, see eqn 8. In particular, we can see that (G=c=1)
  # kappaA_ell = lambdaA_ell * (2*ell-1)!! * XA^4 * (XB)
  kappaA_2 = lambdaA*3.*pow(XA,4)*(1.-XA)
  kappaB_2 = lambdaB*3.*pow(XB,4)*(1.-XB)

  if(lll==2 and mmm==2):
    newtlm = 1.
    tidelm = kappaA_2*(XA/XB+3.)*(1.+beta22_1A*v2) \
             + kappaB_2*(XB/XA+3.)*(1.+beta22_1B*v2)
  elif(lll==2 and mmm==1):
    newtlm = 1j/3.*v
    tidelm = - kappaA_2*(4.5-6.*XA) \
             + kappaB_2*(4.5-6.*XB)
  elif(lll==3 and mmm==3):
    newtlm = (-3./4.*1j*np.sqrt(15./14.))*v
    tidelm = - kappaA_2*6.*(1.-XA) \
             + kappaB_2*6.*(1.-XB)
  elif(lll==3 and mmm==1):
    newtlm = (1j/12./np.sqrt(14.))*v
    tidelm = - kappaA_2*6.*(1.-XA) \
             + kappaB_2*6.*(1.-XB)
  else:
    return 0.

  tidelm *= v10

  return abs(newt*newtlm*tidelm)

################## TaylorT2 PN Tidal Coefficients

## T2 Timing Terms ##

def PNT2QM_Tv4(XA,chiA):
  """ TaylorT2 2PN Quadrupole Moment Coefficient, v^4 Timing Term. 

   XA = mass fraction of object
   chiA = dimensionless spin of object """

  return -10.*XA*XA*chiA*chiA

def PNT2Tidal_Tv10(XA):
  """ TaylorT2 0PN Quadrupolar Tidal Coefficient, v^10 Timing Term. 
   XA = mass fraction of object """

  return 288-264*XA

def PNT2Tidal_Tv12(XA):
  """ TaylorT2 1PN Quadrupolar Tidal Coefficient, v^12 Timing Term. 
   XA = mass fraction of object """

  XATo2nd = XA*XA
  XATo3rd = XATo2nd*XA
  return (3179)/(4)-(919*XA)/(4)-(1143*XATo2nd)/(2) + 65*XATo3rd

def PNT2Tidal_Tv13(XA,chiA=0,chiB=0):
  """ TaylorT2 1.5PN Quadrupolar Tidal Coefficient, v^13 Timing Term. 

   XA = mass fraction of object
   chiA = aligned spin-orbit component of object
   chiB = aligned spin-orbit component of companion object """

  XATo2nd = XA*XA
  XATo3rd = XATo2nd*XA
  return -576*np.pi +(2496*np.pi*XA)/(5)+(324*XA+12*XATo2nd \
    - (1096*XATo3rd)/(5))*chiA+(588-(6616*XA)/(5)+ (4772*XATo2nd)/(5) \
    - (1096*XATo3rd)/(5))*chiB

def PNT2Tidal_Tv14(XA,chiA=0,chiB=0,AqmA=0,AqmB=0,alpha2PNT=0):
  """ TaylorT2 2PN Quadrupolar Tidal Coefficient, v^14 Timing Term. 

   XA = mass fraction of object
   chiA = aligned spin-orbit component of object
   chiB = aligned spin-orbit component of companion object
   AqmA = dimensionless spin-induced quadrupole moment of object
   AqmB = dimensionless spin-induced quadrupole moment of companion object
   alpha2PNT = 2PN Quadrupole Tidal Flux coefficient """

  XATo2nd = XA*XA
  XATo3rd = XATo2nd*XA
  XATo4th = XATo3rd*XA
  XATo5th = XATo4th*XA
  return (70312133/21168)+(4*alpha2PNT)/(3) - (147794303*XA)/(127008) \
    - (20905*XATo2nd)/(28) - (432193*XATo3rd)/(504)-(5848*XATo4th)/(9) \
    + (857*XATo5th)/(3) + (-(639*XATo2nd)/(2)+(525*XATo3rd)/(2) \
    + AqmA*(-312*XATo2nd+256*XATo3rd))*chiA*chiA \
    + (-609*XA+1108*XATo2nd-499*XATo3rd)*chiA*chiB \
    + (-(639)/(2)+(1803*XA)/(2)-(1689*XATo2nd)/(2) + (525*XATo3rd)/(2) \
    + AqmB*(-312+880*XA-824*XATo2nd+256*XATo3rd))*chiB*chiB

def PNT2Tidal_Tv15(XA,chiA=0,chiB=0):
  """ TaylorT2 2.5PN Quadrupolar Tidal Coefficient, v^15 Timing Term. 

   XA = mass fraction of object
   chiA = aligned spin-orbit component of object
   chiB = aligned spin-orbit component of companion object """

  XATo2nd = XA*XA
  XATo3rd = XATo2nd*XA
  XATo4th = XATo3rd*XA
  XATo5th = XATo4th*XA
  return -(241295*np.pi )/(98)+(216921*np.pi*XA)/(98)-(7528*np.pi*XATo2nd)/(7) \
    + (7142*np.pi*XATo3rd)/(7) + ((101949*XA)/(49)+(48875*XATo2nd)/(294) \
    - (78373*XATo3rd)/(147)-(3417*XATo4th)/(7) - (2574*XATo5th)/(7))*chiA \
    + ((637447)/(147)-(1026647*XA)/(98)+(931999*XATo2nd)/(98) \
    - (713938*XATo3rd)/(147)+(12977*XATo4th)/(7) - (2574*XATo5th)/(7))*chiB

def PNT2TidalOcto_Tv14(XA,beta0PNT=0):
  """ TaylorT2 0PN Octopolar Tidal Coefficient, v^14 Timing Term. 

   XA = mass fraction of object
   beta0PNT = 0PN Octopole Tidal Flux coefficient """

  return (4)/(3)*(520+beta0PNT)-(2080*XA)/(3)

def PNT2TidalOcto_Tv16(XA,beta0PNT=0,beta1PNT=0):
  """ TaylorT2 1PN Octopolar Tidal Coefficient, v^16 Timing Term. 

   XA = mass fraction of object
   beta0PNT = 0PN Octopole Tidal Flux coefficient
   beta1PNT = 1PN Octopole Tidal Flux coefficient """

  XATo2nd = XA*XA
  XATo3rd = XATo2nd*XA
  XATo4th = XATo3rd*XA
  XATo5th = XATo4th*XA
  return (33440)/(21)+(995*beta0PNT)/(168)+beta1PNT + ((24670)/(7) \
    + (17*beta0PNT)/(3))*XA-(17/6)*(1825+2*beta0PNT)*XATo2nd+(325*XATo3rd)/(6)

def PNT2TidalOcto_Tv17(XA,chiA=0,chiB=0,beta0PNT=0):
  """ TaylorT2 1.5PN Octopolar Tidal Coefficient, v^17 Timing Term. 

   XA = mass fraction of object
   chiA = aligned spin-orbit component of object
   chiB = aligned spin-orbit component of companion object
   beta0PNT = 0PN Octopole Tidal Flux coefficient """

  XATo2nd = XA*XA
  XATo3rd = XATo2nd*XA
  return -(64)/(9)*(260+beta0PNT)*np.pi + (16640*np.pi*XA)/(9) \
    + ((20)/(9)*(260+3*beta0PNT)*XA+(16)/(27)*(195+7*beta0PNT)*XATo2nd \
    - (2080*XATo3rd)/(3))*chiA + ((4)/(27)*(8580+73*beta0PNT) \
    - (4/27)*(21840+101*beta0PNT)*XA+(16)/(27)*(4485+7*beta0PNT)*XATo2nd \
    + (-2080*XATo3rd)/(3))*chiB

def PNT2TidalOcto_Tv18(XA,chiA=0,chiB=0,AqmA=0,AqmB=0,beta0PNT=0,beta1PNT=0, \
        beta2PNT=0):
  """ TaylorT2 2PN Octopolar Tidal Coefficient, v^18 Timing Term. 

   XA = mass fraction of object
   chiA = aligned spin-orbit component of object
   chiB = aligned spin-orbit component of companion object
   AqmA = dimensionless spin-induced quadrupole moment of object
   AqmB = dimensionless spin-induced quadrupole moment of companion object
   beta0PNT = 0PN Octopole Tidal Flux coefficient
   beta1PNT = 1PN Octopole Tidal Flux coefficient
   beta2PNT = 2PN Octopole Tidal Flux coefficient """

  XATo2nd = XA*XA
  XATo3rd = XATo2nd*XA
  XATo4th = XATo3rd*XA
  XATo5th = XATo4th*XA
  return (35111473)/(3969)+(6080015*beta0PNT)/(254016)+ (199*beta1PNT)/(42) \
    + (4*beta2PNT)/(5)+((68598877)/(7938)  + (16999*beta0PNT)/(840) \
    + (68*beta1PNT)/(15))*XA + ((7215185-16319*beta0PNT \
    - 11424*beta1PNT)*XATo2nd)/(2520) + (-(154907)/(8) \
    - (2477*beta0PNT)/(90))*XATo3rd+(-(3040/3)+(2477*beta0PNT)/(180))*XATo4th \
    + (455*XATo5th)/(18) + ((-858-(57*beta0PNT)/(10))*XATo2nd \
    + 858*XATo3rd+AqmA*((-832 - 28*beta0PNT/5)*XATo2nd \
    + 832*XATo3rd))*chiA*chiA \
    + ((-1612-11*beta0PNT)*XA+(3224 \
    + 11*beta0PNT)*XATo2nd+(-1612*XATo3rd))*chiA*chiB \
    + (-858-(57*beta0PNT)/(10)+(2574+57*beta0PNT/5)*XA \
    + (-2574-(57*beta0PNT)/(10))*XATo2nd + 858*XATo3rd \
    + AqmB*(-832-28/5*beta0PNT + (2496+56/5*beta0PNT)*XA \
    - (4/5)*(3120+7*beta0PNT)*XATo2nd + 832*XATo3rd))*chiB*chiB

def PNT2TidalOcto_Tv19(XA,chiA=0,chiB=0,beta0PNT=0,beta1PNT=0):
  """ TaylorT2 2.5PN Octopolar Tidal Coefficient, v^19 Timing Term. 

   XA = mass fraction of object
   chiA = aligned spin-orbit component of object
   chiB = aligned spin-orbit component of companion object
   beta0PNT = 0PN Octopole Tidal Flux coefficient
   beta1PNT = 1PN Octopole Tidal Flux coefficient """

  XATo2nd = XA*XA
  XATo3rd = XATo2nd*XA
  XATo4th = XATo3rd*XA
  XATo5th = XATo4th*XA
  return -(1)/(462)*(2604260+17705*beta0PNT+2688*beta1PNT)*np.pi \
    - 2/231*(516675 + beta0PNT)*np.pi*XA + (2)/(33)*(90640 \
    + 241*beta0PNT)*np.pi*XATo2nd + (152360*np.pi*XATo3rd)/(33) \
    + ((1)/(924)*(3242620+49273*beta0PNT+5040*beta1PNT)*XA \
    + ((4691970+82637*beta0PNT+4704*beta1PNT)*XATo2nd)/(1386) \
    + (1)/(693)*(-221335+91*beta0PNT)*XATo3rd \
    - (5)/(99)*(98807+254*beta0PNT)*XATo4th - (156910*XATo5th)/(99))*chiA \
    + ((23393100+277897*beta0PNT+24528*beta1PNT)/(2772) \
    + ((-44089360-337219*beta0PNT-33936*beta1PNT)*XA)/(2772) \
    + ((9394160-23497*beta0PNT+4704*beta1PNT)*XATo2nd)/(1386) \
    + ((-272870)/(231)+(563*beta0PNT)/(11))*XATo3rd \
    - (5)/(99)*(-68399+254*beta0PNT)*XATo4th-(156910*XATo5th)/(99))*chiB

## T2 Phasing Terms ##

def PNT2QM_Pv4(XA,chiA):
  """ TaylorT2 2PN Quadrupole Moment Coefficient, v^4 Phasing Term. 

   XA = mass fraction of object
   chiA = dimensionless spin of object """

  return -25.*XA*XA*chiA*chiA

def PNT2Tidal_Pv10(XA):
  """ TaylorT2 0PN Quadrupolar Tidal Coefficient, v^10 Phasing Term.

   XA = mass fraction of object """

  return 72-66*XA

def PNT2Tidal_Pv12(XA):
  """ TaylorT2 1PN Quadrupolar Tidal Coefficient, v^12 Phasing Term. 

   XA = mass fraction of object """

  XATo2nd = XA*XA
  XATo3rd = XATo2nd*XA
  return (15895)/(56)-(4595*XA)/(56) - (5715*XATo2nd)/(28)+(325*XATo3rd)/(14)

def PNT2Tidal_Pv13(XA,chiA=0,chiB=0):
  """ TaylorT2 1.5PN Quadrupolar Tidal Coefficient, v^13 Phasing Term. 

   XA = mass fraction of object
   chiA = aligned spin-orbit component of object
   chiB = aligned spin-orbit component of companion object """

  XATo2nd = XA*XA
  XATo3rd = XATo2nd*XA
  return -225*np.pi +195*np.pi*XA + ((2025*XA)/(16)+(75*XATo2nd)/(16) \
    - (685*XATo3rd)/(8))*chiA+((3675)/(16) - (4135*XA)/(8)+(5965*XATo2nd)/(16) \
    - (685*XATo3rd)/(8))*chiB

def PNT2Tidal_Pv14(XA,chiA=0,chiB=0,AqmA=0,AqmB=0,alpha2PNT=0):
  """ TaylorT2 2PN Quadrupolar Tidal Coefficient, v^14 Phasing Term. 

   XA = mass fraction of object
   chiA = aligned spin-orbit component of object
   chiB = aligned spin-orbit component of companion object
   AqmA = dimensionless spin-induced quadrupole moment of object
   AqmB = dimensionless spin-induced quadrupole moment of companion object
   alpha2PNT = 2PN Quadrupole Tidal Flux coefficient """

  XATo2nd = XA*XA
  XATo3rd = XATo2nd*XA
  XATo4th = XATo3rd*XA
  XATo5th = XATo4th*XA
  return (351560665)/(254016)+(5*alpha2PNT)/(9) - (738971515*XA)/(1524096) \
    - (104525*XATo2nd)/(336) - (2160965*XATo3rd)/(6048)-(7310*XATo4th)/(27) \
    + (4285*XATo5th)/(36) + (-(1065*XATo2nd)/(8)+(875*XATo3rd)/(8) \
    + AqmA*(-130*XATo2nd+(320*XATo3rd)/(3)))*chiA*chiA \
    + (-(1015*XA)/(4)+(1385*XATo2nd)/(3) - (2495*XATo3rd)/(12))*chiA*chiB \
    + (-(1065)/(8)+(3005*XA)/(8)-(2815*XATo2nd)/(8) + (875*XATo3rd)/(8) \
    + AqmB*(-130+(1100*XA)/(3)-(1030*XATo2nd)/(3)+(320*XATo3rd)/(3)))*chiB*chiB

def PNT2Tidal_Pv15(XA,chiA=0,chiB=0):
  """ TaylorT2 2.5PN Quadrupolar Tidal Coefficient, v^15 Phasing Term. 

   XA = mass fraction of object
   chiA = aligned spin-orbit component of object
   chiB = aligned spin-orbit component of companion object """

  XATo2nd = XA*XA
  XATo3rd = XATo2nd*XA
  XATo4th = XATo3rd*XA
  XATo5th = XATo4th*XA
  return -(241295*np.pi)/(224)+(216921*np.pi*XA)/(224)-(941*np.pi*XATo2nd)/(2) \
    + (3571*np.pi*XATo3rd)/(8) + ((101949*XA)/(112)+(48875*XATo2nd)/(672) \
    - (78373*XATo3rd)/(336)-(3417*XATo4th)/(16) - (1287*XATo5th)/(8))*chiA \
    + ((637447)/(336)-(1026647*XA)/(224)+(931999*XATo2nd)/(224) \
    - (356969*XATo3rd)/(168)+(12977*XATo4th)/(16) - (1287*XATo5th)/(8))*chiB

def PNT2TidalOcto_Pv14(XA,beta0PNT=0):
  """ TaylorT2 0PN Octopolar Tidal Coefficient, v^14 Phasing Term. 

   XA = mass fraction of object
   beta0PNT = 0PN Octopole Tidal Flux coefficient """

  return (5)/(9)*(520+beta0PNT)-(2600*XA)/(9)

def PNT2TidalOcto_Pv16(XA,beta0PNT=0,beta1PNT=0):
  """ TaylorT2 1PN Octopolar Tidal Coefficient, v^16 Phasing Term. 

   XA = mass fraction of object
   beta0PNT = 0PN Octopole Tidal Flux coefficient
   beta1PNT = 1PN Octopole Tidal Flux coefficient """

  XATo2nd = XA*XA
  XATo3rd = XATo2nd*XA
  XATo4th = XATo3rd*XA
  XATo5th = XATo4th*XA
  return (5*(267520+995*beta0PNT+168*beta1PNT))/(1848) \
    + (5/231)*(74010 + 119*beta0PNT)*XA \
    - (85)/(66)*(1825+2*beta0PNT)*XATo2nd + (1625*XATo3rd)/(66)

def PNT2TidalOcto_Pv17(XA,chiA=0,chiB=0,beta0PNT=0):
  """ TaylorT2 1.5PN Octopolar Tidal Coefficient, v^17 Phasing Term. 

   XA = mass fraction of object
   chiA = aligned spin-orbit component of object
   chiB = aligned spin-orbit component of companion object
   beta0PNT = 0PN Octopole Tidal Flux coefficient """

  XATo2nd = XA*XA
  XATo3rd = XATo2nd*XA
  return -(10)/(3)*(260+beta0PNT)*np.pi + (2600*np.pi*XA)/(3) \
    + ((25)/(24)*(260+3*beta0PNT)*XA+(5)/(18)*(195+7*beta0PNT)*XATo2nd \
    - 325*XATo3rd)*chiA + ((5)/(72)*(8580+73*beta0PNT) \
    - (5/72)*(21840 + 101*beta0PNT)*XA + (5)/(18)*(4485+7*beta0PNT)*XATo2nd \
    - 325*XATo3rd)*chiB

def PNT2TidalOcto_Pv18(XA,chiA=0,chiB=0,AqmA=0,AqmB=0,beta0PNT=0,beta1PNT=0, \
        beta2PNT=0):
  """ TaylorT2 2PN Octopolar Tidal Coefficient, v^18 Phasing Term. 

   XA = mass fraction of object
   chiA = aligned spin-orbit component of object
   chiB = aligned spin-orbit component of companion object
   AqmA = dimensionless spin-induced quadrupole moment of object
   AqmB = dimensionless spin-induced quadrupole moment of companion object
   beta0PNT = 0PN Octopole Tidal Flux coefficient
   beta1PNT = 1PN Octopole Tidal Flux coefficient
   beta2PNT = 2PN Octopole Tidal Flux coefficient """

  XATo2nd = XA*XA
  XATo3rd = XATo2nd*XA
  XATo4th = XATo3rd*XA
  XATo5th = XATo4th*XA
  return (25*(2247134272+6080015*beta0PNT+1203552*beta1PNT))/(13208832) \
    + (5*beta2PNT)/(13)+((1714971925)/(412776) + (84995*beta0PNT)/(8736) \
    + (85*beta1PNT)/(39))*XA - (5*(-7215185+16319*beta0PNT \
    + 11424*beta1PNT)*XATo2nd)/(26208) \
    - (5*(6970815+9908*beta0PNT)*XATo3rd)/(3744) \
    + (5*(-182400+2477*beta0PNT)*XATo4th)/(1872)+(875*XATo5th)/(72) \
    + (-(15/104)*(2860+19*beta0PNT)*XATo2nd + (825*XATo3rd)/(2) \
    + AqmA*((-400 - (35*beta0PNT)/(13))*XATo2nd+400*XATo3rd))*chiA*chiA \
    + ((-775-(275*beta0PNT)/(52))*XA+(1550 + (275*beta0PNT)/(52))*XATo2nd \
    - 775*XATo3rd)*chiA*chiB \
    + (-(15/104)*(2860+19*beta0PNT)+(15)/(52)*(4290 + 19*beta0PNT)*XA \
    - (15/104)*(8580+19*beta0PNT)*XATo2nd + (825*XATo3rd)/(2) \
    + AqmB*(-400-(35*beta0PNT)/(13)+(1200 + (70*beta0PNT)/(13))*XA \
    + (-1200-(35*beta0PNT)/(13))*XATo2nd + 400*XATo3rd))*chiB*chiB

def PNT2TidalOcto_Pv19(XA,chiA=0,chiB=0,beta0PNT=0,beta1PNT=0):
  """ TaylorT2 2.5PN Octopolar Tidal Coefficient, v^19 Phasing Term. 

   XA = mass fraction of object
   chiA = aligned spin-orbit component of object
   chiB = aligned spin-orbit component of companion object
   beta0PNT = 0PN Octopole Tidal Flux coefficient
   beta1PNT = 1PN Octopole Tidal Flux coefficient"""

  XATo2nd = XA*XA
  XATo3rd = XATo2nd*XA
  XATo4th = XATo3rd*XA
  XATo5th = XATo4th*XA
  return -(5*(2604260+17705*beta0PNT+2688*beta1PNT)*np.pi)/(4704) \
    - (5*(516675+1687*beta0PNT)*np.pi*XA)/(1176)+(5)/(168)*(90640 \
    + 241*beta0PNT)*np.pi*XATo2nd + (95225*np.pi*XATo3rd)/(42)  \
    + ((5*(3242620+49273*beta0PNT+5040*beta1PNT)*XA)/(9408) \
    + (5*(4691970+82637*beta0PNT+4704*beta1PNT)*XATo2nd)/(14112) \
    + (5*(-221335+91*beta0PNT)*XATo3rd)/(7056) \
    - (25*(98807+254*beta0PNT)*XATo4th)/(1008) - (392275*XATo5th)/(504))*chiA \
    + ((5*(23393100+277897*beta0PNT+24528*beta1PNT))/(28224) \
    - (5*(44089360+337219*beta0PNT+33936*beta1PNT)*XA)/(28224) \
    + (5*(9394160-23497*beta0PNT+4704*beta1PNT)/(14112))*XATo2nd \
    + (5*(-272870+11823*beta0PNT)*XATo3rd)/(2352) \
    - (25*(-68399+254*beta0PNT)*XATo4th)/(1008)-(392275*XATo5th)/(504))*chiB

################## TaylorT2 Tidal Evolution

def PNT2Tidal(v, q, lambda2A, lambda3A, AqmA, chizA, lambda2B, lambda3B, AqmB, \
        chizB, order=5):
  """ T2 Tidal corrections time and orbital phase of the binary's evolution.

   Inputs:
     v        -- PN expansion parameter, v = x^(1/2) = (M*Omega)^(1/3)
     q        -- mass ratio, defined as larger mass over smaller mass, MA/MB
     lambda2A -- ell=2 dimensionless tidal defomability of larger object
     lambda3A -- ell=3 dimensionless tidal defomability of larger object
     AqmA     -- dimensionless rotationally-induced quadrupole moment of
                 larger object
     chizA    -- spin component aligned with orbital angular momentum of larger
                 object
     lambda2B -- ell=2 dimensionless tidal defomability of smaller object
     lambda3B -- ell=3 dimensionless tidal defomability of smaller object
     AqmB     -- dimensionless rotationally-induced quadrupole moment of
                 smaller object
     chizB    -- spin component aligned with orbital angular momentum of
                 smaller object
     PNOrder  -- order of the PN tidal expansion beyond leading order, must be
                 one of (0,2,3,4,5)
  
   Outputs:
     dt_tid -- correction to the dimensionless time due to the tidal
               deformability as a function of the PN expansion parameter
     dp_tid -- correction to the dimensionless orbital phase due to the tidal
               deformability as a function of the PN expansion parameter
  
   NOTE: this code can utilize both static tides (the lambdas are input as
     a double) or dynamical tides (an array with length same as times array)"""

  # Sanity check the input
  # Either lambda2AB is an array for dynamical tides, or is a single double for
  #   static tides with a nonnegative value
  try:
    if(len(lambda2A)!=len(v)):
      raise IndexError("ERROR: lambda2A must be a float or be an array with "
              "the same length as the times array!")
  except TypeError:
    if(lambda2A<0):
      raise ValueError("ERROR: lambda2A inputs must be positive!")
  try:
    if(len(lambda2B)!=len(v)):
      raise IndexError("ERROR: lambda2B must be a float or be an array with "
              "the same length as the times array!")
  except TypeError:
    if(lambda2B<0):
      raise ValueError("ERROR: lambda2B inputs must be positive!")

  if(q<1):
    raise ValueError("ERROR: Mass ratio must be > 1 by definition used here!")

  if(abs(chizA)>=1 or abs(chizB)>=1):
    raise ValueError("ERROR: Spin must be < 1!")


  # alpha2PNT, beta?PNT are the missing tidal flux coefficients. Since these
  #   are likely to have a small affect on the waveform, default sets it to 0
  alpha2PNT = 0.
  beta0PNT  = 0.
  beta1PNT  = 0.
  beta2PNT  = 0.

  # Mass fractions
  delta = (q-1.)/(q+1.)
  XA = (delta+1.)/2.
  XB = 1.-XA

  if(order not in [0,2,3,4,5]):
    raise ValueError("ERROR: order must be one of [0,2,3,4,5]!")

  v2 = v*v
  v4 = v2*v2
  v5 = v4*v

  XATo4th = XA*XA*XA*XA
  XATo6th = XATo4th*XA*XA

  XBTo4th = XB*XB*XB*XB
  XBTo6th = XBTo4th*XB*XB

  # Quadrupole tide terms (start at 5PN)
  t_tid = lambda2A*XATo4th*PNT2Tidal_Tv10(XA) \
            + lambda2B*XBTo4th*PNT2Tidal_Tv10(XB)
  p_tid = lambda2A*XATo4th*PNT2Tidal_Pv10(XA) \
            + lambda2B*XBTo4th*PNT2Tidal_Pv10(XB)
  if (order>=2):
    t_tid += v2*(lambda2A*XATo4th*PNT2Tidal_Tv12(XA) \
              + lambda2B*XBTo4th*PNT2Tidal_Tv12(XB))
    p_tid += v2*(lambda2A*XATo4th*PNT2Tidal_Pv12(XA) \
              + lambda2B*XBTo4th*PNT2Tidal_Pv12(XB))
  if (order>=3):
    t_tid += v2*v*(lambda2A*XATo4th*PNT2Tidal_Tv13(XA,chizA,chizB) \
              + lambda2B*XBTo4th*PNT2Tidal_Tv13(XB,chizB,chizA))
    p_tid += v2*v*(lambda2A*XATo4th*PNT2Tidal_Pv13(XA,chizA,chizB) \
              + lambda2B*XBTo4th*PNT2Tidal_Pv13(XB,chizB,chizA))
  if (order>=4):
    t_tid += v4*(lambda2A*XATo4th*PNT2Tidal_Tv14(XA,chizA,chizB,AqmA,AqmB,alpha2PNT) \
              + lambda2B*XBTo4th*PNT2Tidal_Tv14(XB,chizB,chizA,AqmB,AqmA,alpha2PNT))
    p_tid += v4*(lambda2A*XATo4th*PNT2Tidal_Pv14(XA,chizA,chizB,AqmA,AqmB,alpha2PNT) \
              + lambda2B*XBTo4th*PNT2Tidal_Pv14(XB,chizB,chizA,AqmB,AqmA,alpha2PNT))
  if (order==5):
    t_tid += v5*(lambda2A*XATo4th*PNT2Tidal_Tv15(XA,chizA,chizB) \
              + lambda2B*XBTo4th*PNT2Tidal_Tv15(XB,chizB,chizA))
    p_tid += v5*(lambda2A*XATo4th*PNT2Tidal_Pv15(XA,chizA,chizB) \
              + lambda2B*XBTo4th*PNT2Tidal_Pv15(XB,chizB,chizA))

  # Octopole tide terms (start at 7PN)
  t_tid += v4*(lambda3A*XATo6th*PNT2TidalOcto_Tv14(XA,beta0PNT) \
            + lambda3B*XBTo6th*PNT2TidalOcto_Tv14(XB,beta0PNT))
  p_tid += v4*(lambda3A*XATo6th*PNT2TidalOcto_Pv14(XA,beta0PNT) \
            + lambda3B*XBTo6th*PNT2TidalOcto_Pv14(XB,beta0PNT))
  if (order>=2):
    t_tid += v5*v*(lambda3A*XATo6th*PNT2TidalOcto_Tv16(XA,beta0PNT,beta1PNT) \
                + lambda3B*XBTo6th*PNT2TidalOcto_Tv16(XB,beta0PNT,beta1PNT))
    p_tid += v5*v*(lambda3A*XATo6th*PNT2TidalOcto_Pv16(XA,beta0PNT,beta1PNT) \
                + lambda3B*XBTo6th*PNT2TidalOcto_Pv16(XB,beta0PNT,beta1PNT))
  if (order>=3):
    t_tid += v5*v2*(lambda3A*XATo6th*PNT2TidalOcto_Tv17(XA,chizA,chizB,beta0PNT) \
                + lambda3B*XBTo6th*PNT2TidalOcto_Tv17(XB,chizB,chizA,beta0PNT))
    p_tid += v5*v2*(lambda3A*XATo6th*PNT2TidalOcto_Pv17(XA,chizA,chizB,beta0PNT) \
                + lambda3B*XBTo6th*PNT2TidalOcto_Pv17(XB,chizB,chizA,beta0PNT))
  if (order>=4):
    t_tid += v4*v4*(lambda3A*XATo6th*PNT2TidalOcto_Tv18(XA,chizA,chizB,AqmA,AqmB,beta0PNT,beta1PNT,beta2PNT) \
                  + lambda3B*XBTo6th*PNT2TidalOcto_Tv18(XB,chizB,chizA,AqmB,AqmA,beta0PNT,beta1PNT,beta2PNT))
    p_tid += v4*v4*(lambda3A*XATo6th*PNT2TidalOcto_Pv18(XA,chizA,chizB,AqmA,AqmB,beta0PNT,beta1PNT,beta2PNT) \
                  + lambda3B*XBTo6th*PNT2TidalOcto_Pv18(XB,chizB,chizA,AqmB,AqmA,beta0PNT,beta1PNT,beta2PNT))
  if (order==5):
    t_tid += v5*v4*(lambda3A*XATo6th*PNT2TidalOcto_Tv19(XA,chizA,chizB,beta0PNT,beta1PNT) \
                + lambda3B*XBTo6th*PNT2TidalOcto_Tv19(XB,chizB,chizA,beta0PNT,beta1PNT))
    p_tid += v5*v4*(lambda3A*XATo6th*PNT2TidalOcto_Pv19(XA,chizA,chizB,beta0PNT,beta1PNT) \
                + lambda3B*XBTo6th*PNT2TidalOcto_Pv19(XB,chizB,chizA,beta0PNT,beta1PNT))

  dt_tid  = -5./(256.*XA*XB)*v2*t_tid
  dt_tid += -5./(256.*XA*XB)/v4*((AqmA-1)*PNT2QM_Tv4(XA,chizA)+(AqmB-1)*PNT2QM_Tv4(XB,chizB))
  dp_tid  = -1./(32.*XA*XB)*v5*p_tid
  dp_tid += -1./(32.*XA*XB)/v*((AqmA-1)*PNT2QM_Pv4(XA,chizA)+(AqmB-1)*PNT2QM_Pv4(XB,chizB))
  return  dt_tid, dp_tid





""" Basic IO functionality for surrogates """

from __future__ import division  # for python 2

__copyright__ = "Copyright (C) 2014 Scott Field and Chad Galley"
__email__     = "sfield@astro.cornell.edu, crgalley@tapir.caltech.edu"
__status__    = "testing"
__author__    = "Scott Field, Chad Galley"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

import numpy as np
import os as os
import h5py
from .parametric_funcs import function_dict as my_funcs
from .new.spline_evaluation import TensorSplineGrid, fast_tensor_spline_eval
from .catalog import get_modelID_from_filename
import collections

surrogate_description = """* Description of tags:
     
    These files control the relationship between data and surrogate evaluations

      surrogate_mode_type -- key to control the relationship between the basis, 
                             given by B below, and the temporal degrees of freedom,
                             given by evaluating parametric fits.
      fit_type_phase      -- key to select parametric fitting function (phase)
      fit_type_amp        -- key to select parametric fitting function (amp)
      parameterization    -- key to select map from q,M to surrogate's parameterization
      affine_map          -- mapped parameter domain to reference interval for fitting? 
                             supported maps: [low,high] -> [-1,1], [0,1], and [low,high]
      fit_type_norm       -- (Optional) key to select parametric fitting function (norm)

* Description of data:

    Surrogate evaluations are carried out by various operations/evaluations on
    these data files

      time_info       -- (i) tuple (dt, tmin, tmax) OR (ii) Nx2 matrix with 
                         times and quadrature weights
      t_units         -- surrogate's temporal units (values = 'TOverMtot', ...)
      surrogate_units -- currently hardcoded to be 'dimensionless'
                         NOTE: error raised if t_units not 'TOverMtot'
                               since surrogate_units and t_units describe the same
                               information as evaluations are currently coded.
                               In principle they *could* be different
      fit_interval    -- min/max values of parameters used for surrogate fitting
      fitparams_amp   -- fitting parameters for waveform amplitude
      fitparams_phase -- fitting parameters for waveform phase
      greedy_points   -- parameters selected by reduced basis greedy algorithm
      B_1/B_2         -- Basis matrices (their meaning is described below)
      eim_indices     -- (Optional) indices of empirical nodes from time series array `t`
      fitparams_norm  -- (Optional) fitting parameters for waveform norm
      V_1/V_2         -- (Optional) Generalized Vandermonde matrix from empirical 
                         interpolation method
      R_1/R_2         -- (Optional) matrix coefficients relating the reduced 
                         basis to the selected waveforms
      spline_knots    -- (Only for fast_spline surrogates) A concatenated list of spline
                         knots in each dimension
      n_spline_knots  -- (Only for fast_spline surrogates) A list of the number of knots
                         in each dimension


* Surrogate data's dependency on surrogate_mode_type:

   Inspecting the _h_sur member function of EvaluateSingleModeSurrogate 
   shows how the surrogate type is used. Briefly,

    (1) 'waveform_basis' -- The columns of B span a linear approximation space
                            for the complex waveform mode such that h ~ np.dot(self.B, h_EIM)
                            In this case, B_1 (B_2) are the real (imaginary) parts
                            of the basis matrix B. eim_indices is a vector of numbers
                            whose length is exactly the number of columns of B.
                            B = B_1 + j*B_2 is the empirical interpolant operator (`B matrix`)


    (2) 'amp_phase_basis' -- The columns of B.real (B.imag) span a linear approximation
                             space for the amplitude (phase) such that 
                             h ~ np.dot(self.B.real, A_EIM) * exp(1j*np.dot(self.B.imag, Phase_EIM) ).
                             In this case, B_1 (B_2) are the amplitude (phase) basis. 
                             eim_indices contains 2 vectors, the first vector's length
                             matches B_1's columns and the second vector's length
                             matches B_2's columns."""


# helper function 
def extract_mode_from_string(mode_str):
  ''' Each mode's data group name is of the form lELL_mEMM. This 
  helper function will extract the numerical values ELL and EMM'''

  splitkk = mode_str.split('_')
  if splitkk[0][0] == 'l' and splitkk[1][0] == 'm':
    ell = int(splitkk[0][1:])
    emm = int(splitkk[1][1:])
  else:
    raise ValueError
  return ell, emm

##############################################
class SurrogateBaseIO:
  """
* Summary (IO base class): 

    Base class for single-mode text-based surrogate format. This class
    organizes a common set of mandatory and optional data files and
    bookkeeping tags. It serves as a base class from which surrogate read
    and write classes will inherit.


"""
  __doc__+=surrogate_description

  # format variable_name_{h5,txt} = {file_name,h5_group}
  # Note: some of these fields are optional

  ## REQUIRED DATA ##

  _surrogate_ID_h5          = 'surrogate_ID' # From local folder name for text surrogates 

  _surrogate_mode_type_txt  = 'surrogate_mode_type.txt'
  _surrogate_mode_type_h5   = 'surrogate_mode_type' # .txt

  _parameterization_txt     = 'parameterization.txt'
  _parameterization_h5      = 'parameterization' # .txt

  _time_info_txt            = 'time_info.txt' # with min, max, dt or (times, quad weights). depricated (11/23/2016)
  _tmin_h5                  = 'tmin' # 
  _tmax_h5                  = 'tmax' # Pass tmin, tmax and dt, or... 
  _dt_h5                    = 'dt'   #

  _times_h5                 = 'times'              # pass a grid of times and quadrature weights 
  _quadrature_weights_h5    = 'quadrature_weights' # 

  _fit_interval_txt         = 'param_fit_interval.txt'
  _fit_min_h5               = 'fit_min' #  _fit_interval_txt = 'param_fit_interval.txt' (both max and min)
  _fit_max_h5               = 'fit_max'

  _B_1_txt                  = 'B_1.txt'
  _B_2_txt                  = 'B_2.txt'
  _B_h5                     = 'B' # B_1.txt
  _B_phase_h5               = 'B_phase' # B_2.txt
  _B_im_h5                  = 'B_im' # B_2.txt


  ## COULD BE OPTIONAL DATA (depends on surrogate type) ###

  _affine_map_txt           = 'affine_map.txt'
  _affine_map_h5            = 'affine_map' # .txt

  _fit_type_phase_txt       = 'fit_type_phase.txt'
  _fit_type_phase_h5        = 'fit_type_phase'   # .txt

  _fit_type_amp_txt         = 'fit_type_amp.txt'
  _fit_type_amp_h5          = 'fit_type_amp' # .txt
  
  _fit_type_re_txt          = 'fit_type_re.txt'
  _fit_type_re_h5           = 'fit_type_re' # .txt

  _fit_type_im_txt          = 'fit_type_im.txt'
  _fit_type_im_h5           = 'fit_type_im' # .txt

  _fit_type_norm_txt        = 'fit_type_norm.txt'
  _fit_type_norm_h5         = 'fit_type_norm' # .txt

  _fitparams_phase_txt      = 'fitparams_phase.txt' # switched from fit_coeff_phase.txt 11/9/2016
  _fitparams_phase_h5       = 'fitparams_phase' #  .txt

  _fitparams_amp_txt        = 'fitparams_amp.txt'   # switched from fit_coeff_amp.txt 11/9/2016
  _fitparams_amp_h5         = 'fitparams_amp'  # .txt
  
  _fitparams_re_h5          = 'fitparams_re'  # .txt
  _fitparams_im_h5          = 'fitparams_im'  # .txt

  _fitparams_norm_txt  = 'fitparams_norm.txt' # switched from fit_coeff_norm.txt 11/9/2016
  _fitparams_norm_h5   = 'fitparams_norm' # .txt


  ## OPTIONAL DATA ##
  _greedy_points_txt   = 'greedy_points.txt'
  _greedy_points_h5    = 'greedy_points' # .txt

  _eim_indices_txt       = 'eim_indices.txt'
  _eim_indices_h5        = 'eim_indices'       # .txt... text: eim_indices contains 2 vectors if amp/phase
    
  _eim_indices_phase_h5  = 'eim_indices_phase' # rolled into eim_indices
  _eim_amp_h5            = 'eim_amp' # text analog? only used to plot 
  _eim_phase_h5          = 'eim_phase' # text analog? only used to plot
    
  _eim_indices_im_h5     = 'eim_indices_im' # rolled into eim_indices
  _eim_re_h5             = 'eim_re' 
  _eim_im_h5             = 'eim_im' 

  _nr_calib_params_h5    = 'nr_calib_params' 

  _V_1_txt             = 'V_1.txt'
  _V_2_txt             = 'V_2.txt'
  _V_h5                = 'V' # V_1.txt V_2.txt

  _R_1_txt             = 'R_1.txt'
  _R_2_txt             = 'R_2.txt'
  _R_h5                = 'R' # R_1.txt R_2.txt

  _t_units_txt         = 't_units.txt'
  _t_units_h5          = 't_units' # .txt



  #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  def __init__(self):

    ### Make list of required data for reading/writing surrogate data ###
    # this is the minimum data. Surrogate evaluations may require more than
    # the minimum
    self.required = ['times', 'B', 'fitparams_amp', 'fitparams_phase',
                     'fit_min', 'fit_max', 'fit_type_amp', 'fit_type_phase',
                     'surrogate_mode_type', 'parameterization']


  #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  def print_required(self):
    """ Print variable names required for importing and exporting surrogate data"""
    
    print("\nGWSurrogate requires data for the following:")
    
    for kk in self.required:
      print("\t"+kk)
    
    pass

  #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  def check_required(self, dict):
    """ Check if input dictionary has the minimum required surrogate data"""
    #keys = dict.keys()
    
    for kk in self.required:
      #if kk not in keys:
      if kk not in dict:
        raise Exception("\nGWSurrogate requires data for "+kk)
    
    return list(dict.keys())



##############################################
class H5Surrogate(SurrogateBaseIO):
  """Load or export a single-mode surrogate in terms of the function's amplitude and phase from HDF5 data format"""

  __doc__ += surrogate_description
  
  #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  def __init__(self, file=None, mode=None, subdir='', closeQ=True):
    SurrogateBaseIO.__init__(self)
    
    ### Get mode name if supplied and check if subdir ends with '/' ###
    if subdir != '':
      if subdir[-1] == '/':
        self.subdir = subdir
        self.mode = subdir[:-1]
      else:
        self.subdir = subdir + '/'
        self.mode = subdir
    else:
      self.subdir = subdir

    # having quick access to ell, m mode will be more efficient in some contexts
    ell, emm = extract_mode_from_string(self.mode)
    self.mode_ell = int(ell)
    self.mode_emm = int(emm)
    
    ### Check file mode if specified ###
    if mode is not None:
      if mode in ['r', 'w', 'r+', 'a']:
        self._mode = mode
      else:
        raise Exception("File mode not recognized. Must be 'r', 'w', 'r+', 'a'.")
    
    ### Check if file is a pointer or path name (string) and open if in 'r' mode ###
    if file is not None:
      self.type = type(file)
      if self.type == str:
        if self._mode == 'r':
          try:
            self.file = h5py.File(file, 'r')
          except:
            pass
        if self._mode == 'w':
          self.file = h5py.File(file, 'w')
      elif self.type == h5py._hl.files.File:
        if self._mode == 'r' or self._mode == 'w':
          self.file = file
    
      ### If mode is 'r' then import surrogate data ###
      if self._mode == 'r':
        self.load_h5(file, subdir=self.subdir, closeQ=closeQ)
      
    pass
  
  #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  def prepare_mode_data(self, data_mode_class):
    """ Prepare a dictionary to export with entries filled from imported single-mode surrogate data"""
    
    # data_mode_class will usually be of the form surrogate_class.single_mode[<mode_key>]
    
    dict = {}
    
    for kk in data_mode_class.keys:
      dict[kk] = data_mode_class.__dict__[kk]
    
    if 'mode' in data_mode_class.__dict__:
      dict['mode'] = data_mode_class.__dict__['mode']
    
    return dict
  
  #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  def chars_to_string(self, chars):
    return "".join(chr(cc) for cc in chars)

  #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  def string_to_chars(self, string):
    return [ord(cc) for cc in string]
  
  #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  def prepare_data(self, data_class):
    """ Prepare a dictionary to export with entries filled from imported surrogate data"""
    return [self.prepare_mode_data(data_class.single_modes[mm]) for mm in data_class.modes]

  #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  def load_h5(self, file, subdir='', closeQ=True):
    
    ### Open file based on input being a filename or file pointer ###
    self.type = type(file)
    if self.type == str:
      try:
        self.file = h5py.File(file, 'r')
      except:
        raise Exception("Cannot open file.")
    elif self.type == h5py._hl.files.File:
      self.file = file
    
    ### Get data keys listing all available surrogate data ###
    if subdir == '':
      self.keys = list(self.file.keys())
    else:
      # Get keys in the given subdirectory
      self.keys = list(self.file[subdir[:-1]].keys())
      
    ### Get surrogateID ####
    name = self.file.filename.split('/')[-1].split('.')[0]
    
    if self._surrogate_ID_h5 in self.keys:
      self.surrogateID = self.chars_to_string(self.file[subdir+self._surrogate_ID_h5][()])
      if self.surrogateID != name:
        print("\n>>> Warning: surrogateID does not have expected name.")
    else:
      surrogateID = get_modelID_from_filename(file.filename)
      if len(surrogateID) == 0 or len(surrogateID) > 1:
        self.surrogateID = None
        print("\n>>> Warning: No surrogate ID found. Could not deduce ID from file")
      else:
        self.surrogateID = surrogateID[0]
        print("\n>>> Found surrogate ID from file name: %s"%self.surrogateID)
    
    ### Get type of basis used to build surrogate 
    # (e.g., basis for complex waveform or for amplitude and phase)
    self.surrogate_mode_type = self.chars_to_string(self.file[subdir+self._surrogate_mode_type_h5][()])
    
    ### Unpack time info ###
    #self.tmin = self.file[subdir+self._tmin_h5][()]
    #self.tmax = self.file[subdir+self._tmax_h5][()]
    #if self._dt_h5 in self.keys:
    #  self.dt = self.file[subdir+self._dt_h5][()]
    #  self.times = np.arange(self.tmin, self.tmax+self.dt, self.dt)
    #  self.quadrature_weights = self.dt * np.ones(self.times.shape)

    if self._dt_h5 in self.keys and self._tmin_h5 in self.keys:
      print(">>> tmin, tmax, dt are depricated as of 11/23/2016.")
      self.tmin = self.file[subdir+self._tmin_h5][()]
      self.tmax = self.file[subdir+self._tmax_h5][()]
      self.dt = self.file[subdir+self._dt_h5][()]
      self.times = np.arange(self.tmin, self.tmax+self.dt, self.dt)
      self.quadrature_weights = self.dt * np.ones(self.times.shape)
    else:
      self.times = self.file[subdir+self._times_h5][:]
      self.tmin  = self.times[0]
      self.tmax  = self.times[-1]
      try:
        self.quadrature_weights = self.file[subdir+self._quadrature_weights_h5][:]
      except:
        self.quadrature_weights = (self.times[1] - self.times[0]) * np.ones(self.times.shape)
        print("\n>>> Warning: Guessing quadrature weights to be identical with %f"%self.quadrature_weights[0])

    #if self._times_h5 not in self.__dict__.keys():
    #  print "\n>>> Warning: No time samples found or generated."
    #if self._quadrature_weights_h5 not in self.__dict__.keys():
    #  print "\n>>> Warning: No quadrature weights found or generated."
    
    if self._t_units_h5 in self.keys:
      self.t_units = self.chars_to_string(self.file[subdir+self._t_units_h5][()])
    else:
      self.t_units = 'TOverMtot'

    ### redundently fill this variable too -- TODO: should have one var only ###
    if(self.t_units == 'TOverMtot'):
      self.surrogate_units = 'dimensionless'
    else:
      raise ValueError('surrogates must be dimensionless')

    ### Greedy points (ordered by RB selection) ###
    if self._greedy_points_h5 in self.keys:
      self.greedy_points = self.file[subdir+self._greedy_points_h5][:]
    else:
      self.greedy_points = None
      print("Cannot load greedy points...OK")
    
    ### Empirical time index (ordered by EIM selection) ###
    if self.surrogate_mode_type == 'amp_phase_basis':
      try:
        self.eim_indices_amp = self.file[subdir+self._eim_indices_h5][:]
        self.eim_indices_phase = self.file[subdir+self._eim_indices_phase_h5][:]
      except KeyError:
        print("Cannot load eim points...OK") 
    elif self.surrogate_mode_type  == 'waveform_basis':
      try:
        self.eim_indices = self.file[subdir+self._eim_indices_h5][:]
      except KeyError:
        print("Cannot load eim points...OK")
    elif self.surrogate_mode_type == 'coorb_waveform_basis':
      try:
        self.eim_indices_re = self.file[subdir+self._eim_indices_h5][:]
        self.eim_indices_im = self.file[subdir+self._eim_indices_im_h5][:]
      except KeyError:
        print("Cannot load eim points...OK")
    else:
      raise ValueError('invalid surrogate type')

    ### Complex B coefficients ###
    if self.surrogate_mode_type == 'amp_phase_basis':
      self.B_1 = self.file[subdir+self._B_h5][:]
      self.B_2 = self.file[subdir+self._B_phase_h5][:]
    elif self.surrogate_mode_type  == 'waveform_basis':
      self.B = self.file[subdir+self._B_h5][:]	
    # changes in code to allow for coorbital mode surrogates
    elif self.surrogate_mode_type == 'coorb_waveform_basis':
      self.B_1 = self.file[subdir+self._B_h5][:]
      self.B_2 = self.file[subdir+self._B_im_h5][:] 
    else:
      raise ValueError('invalid surrogate type')

    ### Information about phase/amp parametric fit ###
    if self._affine_map_h5 in self.keys:
      self.affine_map = self.chars_to_string(self.file[subdir+self._affine_map_h5][()])
    else:
      self.affine_map = 'none'
    if self.surrogate_mode_type  == 'coorb_waveform_basis':
      self.fitparams_re = self.file[subdir+self._fitparams_re_h5][:]
      self.fitparams_im = self.file[subdir+self._fitparams_im_h5][:]
    else:
      self.fitparams_amp = self.file[subdir+self._fitparams_amp_h5][:]
      self.fitparams_phase = self.file[subdir+self._fitparams_phase_h5][:]
    self.fit_min = self.file[subdir+self._fit_min_h5][()]
    self.fit_max = self.file[subdir+self._fit_max_h5][()]
    self.fit_interval = np.array( [self.fit_min, self.fit_max] )

    # NOTE : this part of code is poorly written
    # to ensure minimal code change, I have kept the naming fit_type_amp and fit_type_phase
    # even though fit_type_re and fit_type_im would have been a better choice
    if self.surrogate_mode_type  == 'coorb_waveform_basis':
      self.fit_type_amp = self.chars_to_string(self.file[subdir+self._fit_type_re_h5][()])
      self.fit_type_phase = self.chars_to_string(self.file[subdir+self._fit_type_im_h5][()])
    else:
      self.fit_type_amp = self.chars_to_string(self.file[subdir+self._fit_type_amp_h5][()])
      self.fit_type_phase = self.chars_to_string(self.file[subdir+self._fit_type_phase_h5][()])

    # TODO: node fitting functions need to be generalized to their own class by using gws.new
    if self.fit_type_amp == "fast_spline_real" and self.fit_type_phase == "fast_spline_imag":
      print("Special case: using fast tensor spline for real and imaginary parts instead of amp/phase")
      print("Loading fast tensor spline breakpoints")

      # TODO: promote data fields (e.g. splint_knots) to SurrogateBaseIO data -- but better to use gws.new
      try: # TODO: 1d surrogates should include n_spline_knots in their data
        n_spline_knots = self.file[subdir+'n_spline_knots'][:]
        remaining_spline_knots = self.file[subdir+'spline_knots'][:]
        spline_knots = []
        for n in n_spline_knots:
          spline_knots.append(remaining_spline_knots[:n])
          remaining_spline_knots = remaining_spline_knots[n:]
      except KeyError: # if n_spline_knots does not exist a KeyError is raised. Old 1d surrogate assumed
        spline_knots = [self.file[subdir+'spline_knots'][:]]

      # setup the function which will be used to evaluate splines
      # TODO: unfortunately, this creates a new grid for each mode.
      #       which does not take advantage of grids that are identical
      #       for different modes.
      #       This will also be slow: each mode will need to find the fast spline data for all modes 
      #       without reusing already computed data
      self.ts_grid = TensorSplineGrid(spline_knots)

      # NOTE: this evaluates the real part -- but its called amp to keep with naming convention -- terrible! (TODO)
      def amp_fit_func(coeffs,xvec):
        amp_eval = fast_tensor_spline_eval(xvec,self.ts_grid,coeffs)
        return amp_eval

      self.amp_fit_func = amp_fit_func

      # NOTE: this evaluates the imaginary part -- but its called phase to keep with naming convention -- terrible! (TODO)
      def phase_fit_func(coeffs,xvec):
        phase_eval = fast_tensor_spline_eval(xvec,self.ts_grid,coeffs)
        return phase_eval

      self.phase_fit_func = phase_fit_func

    elif self.fit_type_amp == "spline_1d" and self.fit_type_phase == "spline_1d":
        
      print("Special case: using spline for parametric model at each EI node")
        
      ## NOTE :We need two if statements for the BHPTNRSur1dq1e4 because, the number of knots are not same
      ## in different modes or even for different data pieces within the mode. This is because of the
      ## smoothing we do in fitting. This demands some parts of the code to be re-written.
      ## Furthermore, we use different decomposition for 22 mode and higher order modes.
      ## This should be done differently
      ## =====================================================================================
      if self.surrogateID=="BHPTNRSur1dq1e4":
        if self.surrogate_mode_type  == 'amp_phase_basis': # BHPTNRSur1dq1e4's 22 mode is of type amp_phase_basis
          n_spline_knots_amp = self.file[subdir+'n_spline_knots_amp'][:]
          n_spline_knots_phase = self.file[subdir+'n_spline_knots_phase'][:]

          spline_knots_amp = self.file[subdir+'spline_knots_amp'][:]
          spline_knots_phase = self.file[subdir+'spline_knots_phase'][:]

          # pack necessary data up so the spline function can be called
          # as self.amp_fit_func(self.fitparams_amp[jj,:], x_0)
          num_fits_amp = self.fitparams_amp.shape[0]
          num_fits_phase = self.fitparams_phase.shape[0]
          fitparams_amp = []
          fitparams_phase = []
          degree = int(self.file[subdir+'degree'][:]) # must be int for scipy (> 1.5.2) splev to work
          for i in range(num_fits_amp):
            fitparams_amp.append([spline_knots_amp[i], self.fitparams_amp[i], degree])
          for i in range(num_fits_phase):
            fitparams_phase.append([spline_knots_phase[i], self.fitparams_phase[i], degree])
          # these are ragged structures, so we need to explicitly request dtype=object
          # as it cannot be converted to an arrary with numerical datatypes
          # Instead, it needs to be  saved as an object type, which appears to be an array of pointers
          # to arrays. This could have some efficiency impact.
          # https://stackoverflow.com/questions/29877508/what-does-dtype-object-mean-while-creating-a-numpy-array
          self.fitparams_amp = np.array(fitparams_amp,dtype=object)
          self.fitparams_phase = np.array(fitparams_phase,dtype=object)

          print("num_fits_amp = %i"%(num_fits_amp))
          print("num_fits_phase = %i"%(num_fits_phase))

          self.amp_fit_func   = my_funcs[self.fit_type_amp]
          self.phase_fit_func = my_funcs[self.fit_type_phase]
        elif self.surrogate_mode_type  == 'coorb_waveform_basis':  # BHPTNRSur1dq1e4's higher order modes are coorb type
          n_spline_knots_re = self.file[subdir+'n_spline_knots_re'][:]
          n_spline_knots_im = self.file[subdir+'n_spline_knots_im'][:]

          spline_knots_re = self.file[subdir+'spline_knots_re'][:]
          spline_knots_im = self.file[subdir+'spline_knots_im'][:]

          # pack necessary data up so the spline function can be called
          # as self.amp_fit_func(self.fitparams_amp[jj,:], x_0)
          num_fits_re = self.fitparams_re.shape[0]
          num_fits_im = self.fitparams_im.shape[0]
          fitparams_re = []
          fitparams_im = []
          degree = int(self.file[subdir+'degree'][:]) # must be int for scipy (> 1.5.2) splev to work
          for i in range(num_fits_re):
            fitparams_re.append([spline_knots_re[i], self.fitparams_re[i], degree])
          for i in range(num_fits_im):
            fitparams_im.append([spline_knots_im[i], self.fitparams_im[i], degree])
          # these are ragged structures, so we need to explicitly request dtype=object
          # as it cannot be converted to an arrary with numerical datatypes
          # Instead, it needs to be  saved as an object type, which appears to be an array of pointers
          # to arrays. This could have some efficiency impact.
          # https://stackoverflow.com/questions/29877508/what-does-dtype-object-mean-while-creating-a-numpy-array
          self.fitparams_re = np.array(fitparams_re,dtype=object)
          self.fitparams_im = np.array(fitparams_im,dtype=object)

          print("num_fits_re = %i"%(num_fits_re))
          print("num_fits_im = %i"%(num_fits_im))

          self.re_fit_func   = my_funcs[self.fit_type_amp]
          self.im_fit_func = my_funcs[self.fit_type_phase]
      else: # for all other models, it will take the usual route
        n_spline_knots = self.file[subdir+'n_spline_knots'][:]
        spline_knots = self.file[subdir+'spline_knots'][:]

        # pack necessary data up so the spline function can be called
        # as self.amp_fit_func(self.fitparams_amp[jj,:], x_0)
        num_fits = self.fitparams_amp.shape[0]
        fitparams_amp = []
        fitparams_phase = []
        degree = int(self.file[subdir+'degree'][:]) # must be int for scipy (> 1.5.2) splev to work
        for i in range(num_fits):
          fitparams_amp.append([spline_knots, self.fitparams_amp[i,:], degree])
          fitparams_phase.append([spline_knots, self.fitparams_phase[i,:], degree])
        # these are ragged structures, so we need to explicitly request dtype=object
        # as it cannot be converted to an arrary with numerical datatypes
        # Instead, it needs to be  saved as an object type, which appears to be an array of pointers
        # to arrays. This could have some efficiency impact.
        # https://stackoverflow.com/questions/29877508/what-does-dtype-object-mean-while-creating-a-numpy-array
        self.fitparams_amp = np.array(fitparams_amp,dtype=object)
        self.fitparams_phase = np.array(fitparams_phase,dtype=object)

        print("spline knots = %i, num_fits = %i"%(n_spline_knots,num_fits))

        self.amp_fit_func   = my_funcs[self.fit_type_amp]
        self.phase_fit_func = my_funcs[self.fit_type_phase]
    
    else:
      self.amp_fit_func   = my_funcs[self.fit_type_amp]
      self.phase_fit_func = my_funcs[self.fit_type_phase]


    if self._fit_type_norm_h5 in self.keys:
      try:
        self.fitparams_norm = self.file[subdir+self._fitparams_norm_h5][:]
      except KeyError:
      	self.fitparams_norm = None
      	print("setting norm fitparams to None...")
      self.fit_type_norm = self.chars_to_string(self.file[subdir+self._fit_type_norm_h5][()])
      self.norm_fit_func  = my_funcs[self.fit_type_norm]
      self.norms = True
    else:
      self.norms = False
    
    if self._eim_amp_h5 in self.keys:
      self.eim_amp = self.file[subdir+self._eim_amp_h5][:]
    
    if self._eim_phase_h5 in self.keys:
      self.eim_phase = self.file[subdir+self._eim_phase_h5][:]
    
    if self._eim_re_h5 in self.keys:
      self.eim_re = self.file[subdir+self._eim_re_h5][:]
    
    if self._eim_im_h5 in self.keys:
      self.eim_im = self.file[subdir+self._eim_im_h5][:]
    
    ### Transpose matrices if surrogate was built using ROMpy ###
    transposeB = False
    #if not self.surrogate_mode_type == 'amp_phase_basis':
    if self.surrogate_mode_type not in ['amp_phase_basis','coorb_waveform_basis']:
      Bshape = np.shape(self.B)
    
      if Bshape[0] < Bshape[1]:
        transposeB = True
        self.B = np.transpose(self.B)
        self.dim_rb = Bshape[0]
        self.time_samples = Bshape[1]
    
      else:
        self.dim_rb = Bshape[1]
        self.time_samples = Bshape[0]
    else: # TODO: elif... to match other similar control statements
        Bshape = np.shape(self.B_1)
        self.dim_rb = Bshape[0]
        self.time_samples = Bshape[1]
        self.dim_rb_phase = np.shape(self.B_2)[0]

    ### Vandermonde V such that E (orthogonal basis) is E = BV ###
    if self._V_h5 in self.keys:
      self.V = self.file[subdir+self._V_h5][:]
      if transposeB:
        self.V = np.transpose(self.V)
    else:
       self.V = None
    
    ### R matrix such that waveform basis H = ER ###
    if self._R_h5 in self.keys:
      self.R = self.file[subdir+self._R_h5][:]
      if transposeB:
        self.R = np.transpose(self.R)
    else:
      self.R = None
        
    ### Information about surrogate's parameterization ###
    self.parameterization = self.chars_to_string(self.file[subdir+self._parameterization_h5][()])
    self.get_surr_params  = my_funcs[self.parameterization]
    
    if closeQ:
      self.file.close()


    ### check that data has been loaded as 64-bit double (NOTE: not all datasets are checked here)
    ### Why is this important? It was found that using np.float32 can lead to problems,
    ### for example when trying to comput flow, which is sensitive to small changes, round-off
    ### errors can give wildly incorrect values (off by ~10%)
    if not isinstance(self.times[0], np.float64):
      print("Time grid loaded with data type %s. Changing to float64..."%type(self.times[0]))
      self.times = np.array(self.times, dtype=np.float64)
      self.tmin = self.times[0]
      self.tmax = self.times[-1]
    if self.surrogate_mode_type == 'amp_phase_basis':
      if not isinstance(self.B_1[0][0], np.float64):
        print("Basis matrix loaded with data type %s. Changing to float64..."%type(self.B_1[0][0]))
        self.B_1 = np.array(self.B_1, dtype=np.float64)
        self.B_2 = np.array(self.B_2, dtype=np.float64)
    
    elif self.surrogate_mode_type == 'coorb_waveform_basis':
      if not isinstance(self.B_1[0][0], np.float64):
        print("Basis matrix loaded with data type %s. Changing to float64..."%type(self.B_1[0][0]))
        self.B_1 = np.array(self.B_1, dtype=np.float64)
        self.B_2 = np.array(self.B_2, dtype=np.float64)
        
    elif self.surrogate_mode_type  == 'waveform_basis':
      if not isinstance(np.real(self.B[0][0]), np.float64):
        print("Basis matrix loaded with data type %s. Should change to float64...BUT NOT CODED YET!!!"%type(self.B_1[0][0]))
      if self.fit_type_amp == "spline_1d" and self.fit_type_phase == "spline_1d":
        if not isinstance(self.fitparams_amp[0][0][0], np.float64):
          print("Fit parameters loaded with data type %s. This is probably OK. NOT changing to float64..."%type(self.fitparams_amp[0][0][0]))
    #    self.fitparams_amp = np.array(self.fitparams_amp, dtype=np.float64)
    #    self.fitparams_phase = np.array(self.fitparams_phase, dtype=np.float64)
    
    elif self.surrogate_mode_type  == 'coorb_waveform_basis':
      if self.fit_type_re == "spline_1d" and self.fit_type_im == "spline_1d":
        if not isinstance(self.fitparams_re[0][0][0], np.float64):
          print("Fit parameters loaded with data type %s. This is probably OK. NOT changing to float64..."%type(self.fitparams_re[0][0][0]))
    pass
    
  
  #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  def write_h5(self, dict, path=None, subdir='', closeQ=True):
    """ Export surrogate data in standard format.
    
    Input:
    ======
      dict -- Dictionary or list of dictionaries 
              containing surrogate data to export.
      
    NOTE: Run print_required() to print the list of 
    the minimum data required by GWSurrogate.
    """
    
    ### Check if path is a filename or pointer and process accordingly ###
    if path is not None:
      if type(path) is str:
        file = h5py.File(path, 'w')
      elif type(path) == h5py._hl.files.File:
        file = path
    else:
      file = self.file
    
    ### Check if dict is for a single mode or multiple modes ###
    if type(dict) is list:
      num_modes = len(dict)
    else:
      num_modes = 1
     
    ### Check that the minimum required surrogate data is given ###
    if num_modes == 1:
      keys = [self.check_required(dict)]
    else:
      keys = [self.check_required(dd) for dd in dict]
    
    ### Export surrogate data to HDF5 file ###
    for ii in range(num_modes):
      
      # Create a group that will house a set of surrogate data for a single mode
      if num_modes == 1:
        data_to_write = dict
        if subdir == '':
          group = file.create_group(data_to_write['mode'])
        else:
          group = file.create_group(subdir)
      else:
        data_to_write = dict[ii]
        group = file.create_group(data_to_write['mode'])
      
      # Write single-mode surrogate data to file, excluding the 'mode' information
      for kk in keys[ii]:

        print("key = %s"%kk)

        if kk not in ['mode', 'surrogate_ID', 'B1_spline_params', 'B2_spline_params']:
            
          dtype = type(data_to_write[kk])
          if dtype is str:
            chars = self.string_to_chars(data_to_write[kk])
            group.create_dataset(kk, data=chars, dtype='int')
          elif dtype is np.ndarray:
            group.create_dataset(kk, data=data_to_write[kk], dtype=data_to_write[kk].dtype, compression='gzip')
          elif isinstance(data_to_write[kk], collections.Callable):
            print("key %s is a function. Not writing to h5 file"%kk)
          elif data_to_write[kk] is None:
            print("key %s has a value of none. Not writing to h5 file"%kk)
          else:
            group.create_dataset(kk, data=data_to_write[kk], dtype=type(data_to_write[kk]))

        elif kk=='surrogate_ID':
          name = file.filename.split('/')[-1].split('.')[0]
          group.create_dataset('surrogate_ID', data=self.string_to_chars(name), dtype='int')
          #group.create_dataset('surrogate_ID', data=self.string_to_chars(data_to_write[kk]), dtype='int')
    
    ### Close file, if requested ###
    if closeQ:
      file.close()
    
    pass


##############################################
class TextSurrogateRead(SurrogateBaseIO):
  """Load single-mode, text-based surrogate"""
  __doc__+=surrogate_description

  #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  def __init__(self, sdir):
    """initialize single-mode surrogate defined from text files 
       located in directory sdir. """

    surrogate_load_info = '' # add to string, display after loading

    ### sdir is defined to be the surrogate's ID ###
    self.surrogateID = sdir

    ### type of surrogate (for harmonic mode) ###
    self.surrogate_mode_type = \
      self.get_string_key(sdir+self._surrogate_mode_type_txt)

    ### Surrogate's sampling rate and mass ratio (for fits) ###
    self.time_info    = np.loadtxt(sdir+self._time_info_txt)
    self.fit_interval = np.loadtxt(sdir+self._fit_interval_txt)
    self.fit_min = self.fit_interval[0]
    self.fit_max = self.fit_interval[1]

    ### unpack time info ###
    # NOTE: models stored in dimensionless T/M, whereas basis may be
    # constructed in T. Basis not necessarily ortho wrt dt (see basis.ipynb)
    # newer surrogate models use basis defined on stored time grid
    if(self.time_info.size == 3):
      self.dt      = self.time_info[2]
      self.tmin    = self.time_info[0]
      self.tmax    = self.time_info[1]

      self.times              = np.arange(self.tmin, self.tmax+self.dt, self.dt)
      self.quadrature_weights = self.dt * np.ones(self.times.shape)
    else:
      self.times              = time_info[:,0]
      self.quadrature_weights = time_info[:,1]
      self.tmin               = np.min(self.times)
      self.tmax               = np.max(self.times)

    self.time_samples = self.times.shape[0]
    self.dt           = self.times[1] - self.times[0]

    try:
      self.t_units = self.get_string_key(sdir+self._t_units_txt)
    except IOError:
      self.t_units = 'TOverMtot'

    ### redundently fill this variable too -- TODO: should have one var only ###
    if(self.t_units == 'TOverMtot'):
      self.surrogate_units = 'dimensionless'
    else:
      raise ValueError('surrogates must be dimensionless')

    ### Complex B coefficients - set ndim=2 in case only 1 basis vector ###
    B_1    = np.loadtxt(sdir+self._B_1_txt,ndmin=2)
    B_2    = np.loadtxt(sdir+self._B_2_txt,ndmin=2)

    ### Consistency check that self.time_samples = B_X.shape[0] ###
    if(self.time_samples != B_1.shape[0] or
       self.time_samples != B_2.shape[0]):
      print(self.time_samples)
      print(B_2.shape[0])
      raise ValueError('temporal and basis dimension mismatch')


    ### Cases handled by in evaluation code ###
    if self.surrogate_mode_type  == 'amp_phase_basis':
      self.B_1 = B_1
      self.B_2 = B_2
      self.B   = None
      self.modeled_data  = 2 # amp, phase data
      self.fits_required = 2 # amp, phase fits
    elif self.surrogate_mode_type  == 'waveform_basis':
      self.B   = B_1 + (1j)*B_2
      self.B_1 = None
      self.B_2 = None
      self.modeled_data  = 1 # complexified waveform data
      self.fits_required = 2 # amp, phase fits
    elif self.surrogate_mode_type  == 'coorb_waveform_basis':
      raise ValueError('coorb_waveform_basis model not supported for text datafiles')
      #self.B   = B_1 + (1j)*B_2
      #self.B_1 = None
      #self.B_2 = None
      #self.modeled_data  = 1 # complexified waveform data
      #self.fits_required = 2 # amp, phase fits
    else:
      raise ValueError('invalid surrogate type')

    ### Information about phase/amp parametric fits ###
    self.fitparams_phase = np.loadtxt(sdir+self._fitparams_phase_txt,ndmin=2)
    self.fitparams_amp   = np.loadtxt(sdir+self._fitparams_amp_txt,ndmin=2)

    self.affine_map      = self.get_string_key(sdir+self._affine_map_txt)

    self.fit_type_phase  = self.get_string_key(sdir+self._fit_type_phase_txt)
    self.fit_type_amp    = self.get_string_key(sdir+self._fit_type_amp_txt)

    self.phase_fit_func = my_funcs[self.fit_type_phase]
    self.amp_fit_func   = my_funcs[self.fit_type_amp]

    ### Information about surrogate's parameterization ###
    self.parameterization = self.get_string_key(sdir+self._parameterization_txt)
    self.get_surr_params  = my_funcs[self.parameterization]


    ### Load optional parameters if they exist ###

    ### Vandermonde V such that E (orthogonal basis) is E = BV ###
    try:
      V_1    = np.loadtxt(sdir+self._V_1_txt)
      V_2    = np.loadtxt(sdir+self._V_2_txt)
      self.V = V_1 + (1j)*V_2
    except IOError:
      surrogate_load_info +='Vandermonde not found, '
      self.V = False


    ### greedy points (ordered by RB selection) ###
    try:
      self.greedy_points = np.loadtxt(sdir+self._greedy_points_txt)
    except IOError:
      surrogate_load_info += 'Greedy points not found, '
      self.greedy_points = False

    ### R matrix such that waveform basis H = ER ###
    try:
      R_1    = np.loadtxt(sdir+self._R_1_txt)
      R_2    = np.loadtxt(sdir+self._R_2_txt)
      self.R = R_1 + (1j)*R_2
    except IOError:
      surrogate_load_info += 'R matrix not found, '
      self.R = False

    try: 
      self.fitparams_norm = np.loadtxt(sdir+self._fitparams_norm_txt)
      self.fit_type_norm  = self.get_string_key(sdir+self._fit_type_norm_txt)
      self.norm_fit_func  = my_funcs[self.fit_type_norm]
      self.norms = True
    except IOError:
      surrogate_load_info += 'Norm fits not found, '
      self.norms = False

    ### empirical time index (ordered by EIM selection) ###
    try:
      self.eim_indices = np.loadtxt(sdir+self._eim_indices_txt,dtype=int)
    except IOError:
      surrogate_load_info += 'EIM indices not found.'
      self.eim_indices = False

    #print surrogate_load_info #Q: should we display this?

  #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  def get_string_key(self,fname):
    """ return a single word string from file """

    fp = open(fname,'r')
    keyword = fp.readline()
    if '\n' in keyword:
      return keyword[0:-1]
    else:
      return keyword


##############################################
class TextSurrogateWrite(SurrogateBaseIO):
  """Export single-mode, text-based surrogate"""
  __doc__+=surrogate_description

  #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  def __init__(self, sdir):
    """open single-mode surrogate, to be located in directory sdir, for writing"""

    if( sdir[-1:] != '/' ):
      raise Exception("path name should end in /")
    try:
      os.mkdir(sdir)
      print("Successfully created a surrogate directory...use write_text to export your surrogate!")
    except OSError:
      print("Could not create a surrogate directory. Not ready to export, please try again.")

    ### sdir is defined to be the surrogate's ID ###
    self.surrogateID = sdir


  #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  def write_text(self, time_info, B, eim_indices, greedy_points, \
                 fit_interval, affine_map, \
                 fitparams_amp, fitparams_phase, fitparams_norm, V, R, \
                 fit_type_phase,\
                 fit_type_amp, fit_type_norm,parameterization,\
                 surrogate_mode_type):
    """ Write surrogate data (text) in standard format."""
		

    # TODO: flag to zip folder and save full time series

    ### pack mass ratio interval (for fits) and time info ###
    self._np_savetxt_safe(self.surrogateID+self._fit_interval_txt,fit_interval)
    self._np_savetxt_safe(self.surrogateID+self._time_info_txt,time_info)
    self._np_savetxt_safe(self.surrogateID+self._greedy_points_txt,\
                         greedy_points,fmt='%2.16f')
    self._np_savetxt_safe(self.surrogateID+self._eim_indices_txt,\
                         eim_indices,fmt='%i')
    self._np_savetxt_safe(self.surrogateID+self._B_1_txt,B.real)
    self._np_savetxt_safe(self.surrogateID+self._B_2_txt,B.imag)
    self._np_savetxt_safe(self.surrogateID+self._fitparams_phase_txt,\
                         fitparams_phase)
    self._np_savetxt_safe(self.surrogateID+self._fitparams_amp_txt,\
                         fitparams_amp)
    #self._np_savetxt_safe(self.surrogateID+self._affine_map_txt,\
    #                     np.array([int(affine_map)]),fmt='%i')
    self._np_savetxt_safe(self.surrogateID+self._affine_map_txt,\
                         [affine_map],'%s')
    self._np_savetxt_safe(self.surrogateID+self._V_1_txt,V.real)
    self._np_savetxt_safe(self.surrogateID+self._V_2_txt,V.imag)
    self._np_savetxt_safe(self.surrogateID+self._R_1_txt,R.real)
    self._np_savetxt_safe(self.surrogateID+self._R_2_txt,R.imag)
    self._np_savetxt_safe(self.surrogateID+self._fitparams_norm_txt,\
                         fitparams_norm)
    self._np_savetxt_safe(self.surrogateID+self._fit_type_phase_txt,\
                         [fit_type_phase],'%s')
    self._np_savetxt_safe(self.surrogateID+self._fit_type_amp_txt,\
                         [fit_type_amp],'%s')
    self._np_savetxt_safe(self.surrogateID+self._fit_type_norm_txt,\
                         [fit_type_norm],'%s')
    self._np_savetxt_safe(self.surrogateID+self._parameterization_txt,\
                         [parameterization],'%s')
    self._np_savetxt_safe(self.surrogateID+self._surrogate_mode_type_txt,\
                         [surrogate_mode_type],'%s')

  #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  def _np_savetxt_safe(self,fname, X, fmt='%.18e', delimiter=' ',\
                      newline='\n', header='', footer='', comments='# '):
    """ numpys savetext without overwrites """

    if os.path.isfile(fname):
      raise Exception("file already exists")
    else: 
      np.savetxt(fname,X,fmt=fmt)

# ##############################################
class BHPTNRCalibValues:
  """
  This class reads the NR calibration information from BHPTNRSurrogates
  """
  #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  def __init__(self, file):
    with h5py.File(file, 'r') as f:
        # nr calibration info
        self.alpha_coeffs = {}
        for mode in [(2,2),(3,3),(4,4),(5,5)]:
            self.alpha_coeffs[mode] = f["nr_calib_params/(%d,%d)"%(mode[0],mode[1])]['alpha'][:]
        self.beta_coeffs = f["nr_calib_params/(2,2)"]['beta'][:]
        
    
    





import os

from setuptools import setup, Extension

import numpy

# all extensions here
extmods = []

# build extension 1: python wrapper to gsl's spline function
if os.path.isdir("/opt/local/include"):
    IncDirs = ["/opt/local/include"]
else:
    IncDirs = []

if os.path.isdir("/opt/local/lib"):
    LibDirs = ["/opt/local/lib"]
else:
    LibDirs = []

extmod = Extension(
    "gwsurrogate.spline_interp_Cwrapper._spline_interp",
    include_dirs=IncDirs,
    libraries=["gsl"],
    extra_compile_args=["-std=c99"],
    library_dirs=LibDirs,
    sources=["gwsurrogate/spline_interp_Cwrapper/_spline_interp.c"],
)
extmods.append(extmod)

# build extension 2: precessing utils
extmod = Extension(
    "gwsurrogate.precessing_utils._utils",
    sources=["gwsurrogate/precessing_utils/src/precessing_utils.c"],
    include_dirs=["gwsurrogate/precessing_utils/include", numpy.get_include()],
    language="c",
    extra_compile_args=["-std=c99", "-fPIC", "-O3"],
)
extmods.append(extmod)


# Extract code version from surrogate.py
def read_main_file(key):
    with open("gwsurrogate/surrogate.py") as f:
        for line in f.readlines():
            if key in line:
                return line.split('"')[1]


# define models to be used within pycbc
entries = {
    "pycbc.waveform.td": [
        "GWS-NRHybSur3dq8 = gwsurrogate.pycbc:gws_td_gen",
        "GWS-NRSur7dq4 = gwsurrogate.pycbc:gws_td_gen",
        "GWS-NRHybSur3dq8Tidal = gwsurrogate.pycbc:gws_td_gen",
    ]
}

setup(
    name="gwsurrogate",
    version=read_main_file("__version__"),
    author=read_main_file("__author__"),
    author_email="sfield@umassd.edu",
    packages=[
        "gwsurrogate",
        "gwsurrogate.eval_pysur",
        "gwsurrogate.new",
        # "gwsurrogate.precessing_utils",
        "gwsurrogate.spline_interp_Cwrapper",
    ],
    license="MIT",
    include_package_data=True,
    contributors=[
        # Alphabetical by last name.
        ""
    ],
    description="An easy to use interface to gravitational wave surrogate models",
    long_description_content_type="text/markdown",
    long_description=open("README.md").read(),
    # will start new downloads if these are installed in a non-standard location
    install_requires=[
        "numpy",
        "scipy",
        "h5py",
        "scikit-learn",
        "gwtools",
        "matplotlib",
    ],
    setup_requires=["numpy"],
    classifiers=[
        "Intended Audience :: Other Audience",
        "Intended Audience :: Science/Research",
        "Natural Language :: English",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Mathematics",
        "Topic :: Scientific/Engineering :: Physics",
    ],
    entry_points=entries,
    ext_modules=extmods,
)
